
class Agent
{
public:
	Agent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate)
	{
		agentNumber = agent_number;
		totalUtility = 0.0;
		discountFactorDistribution = discount_factor_distribution;
		
		//initialize the discount factor, then set the real replacement rate
		replacementRate = 1.0;
		UpdateDiscountFactor(0.0);
		replacementRate = replacement_rate;

		for(int i = 0; i < num_agents; i++)
			encounterRate.push_back(0.0);
	}
	
	//updates the discount factor based on the replacement rate
	void UpdateDiscountFactor(double cur_time)
	{
		if(genrand_real1() <= replacementRate)
		{
			discountFactor = discountFactorDistribution->GetNextRand();
			lastReplacementTime = cur_time;
		}
	}
	
	virtual void PrintInfo()
	{
		printf("agent %d:\t$: %f\tdf: %f\tr: %f\n", agentNumber, totalUtility, discountFactor, replacementRate);
	}

	int agentNumber;
	double totalUtility; //total utility gained by the agent so far

	double replacementRate;	//rate at which discount factors are replaced.  1.0 means every round

	double discountFactor;	//current discount factor

	//rate (in [0, 1]) which this agent will play in the role of a1 in an encounter the other agent whose number is represented by the index number
	std::vector<double> encounterRate;

protected:
	double lastReplacementTime; //time at which the last discount factor was set
	
	RandomVariable *discountFactorDistribution;
};


template <class AgentType, class GameType>
class MultiAgentSimulation
{
public:
	//returns true if it has reached the last round, false if not (infinite games always return false)
	virtual bool ExecuteRound(double cur_time) = 0;

	void PrintAgentsInfo()
	{
		for(int i = 0; i < agents->size(); i++)
			(*agents)[i]->PrintInfo();
	}

	void RunSimulation()
	{
		bool still_simulating = true;
		int i;
		for(i = 0; still_simulating; i++)
		{
			printf("time: %d\n", i);
			still_simulating = ExecuteRound(i);
			PrintAgentsInfo();
		}
		printf("time: %d\n", i);
		PrintAgentsInfo();
	}

	void RunSimulation(int num_rounds)
	{
		for(int i = 0; i < num_rounds; i++)
		{
			printf("time: %d\n", i);
			ExecuteRound(i);
			PrintAgentsInfo();
		}
		printf("time: %d\n", num_rounds);
		PrintAgentsInfo();
	}

	void InitializeUniformConnections(RandomVariable *agent_interaction_rate)
	{
		int num_agents = agents->size();
		for(int i = 0; i < num_agents; i++)
			for(int j = 0; j < num_agents; j++)
				(*agents)[i]->encounterRate[j] = agent_interaction_rate->GetNextRand();
	}

	void InitializeSmallWorldConnections(RandomVariable *agent_interaction_rate)
	{
		int num_agents = agents->size();

		//sort of a small world like graph based on probablities
		// similar to the BarabsiAlbert (BA) model
		(*agents)[0]->encounterRate[1] = (*agents)[1]->encounterRate[0] = 1.0;
		double total_weight = 2.0;
		for(int i = 2; i < num_agents; i++)
		{
			//adding new agent i
			double new_total_weight = 0.0;
			for(int j = 0; j < num_agents; j++)
			{
				if(j == i) continue;	//don't connect with self

				//find the weight exhibited by this particular agent
				double weight_of_j = 0.0;
				for(int k = 0; k < i; k++)
					if(k != j)
						weight_of_j += (*agents)[j]->encounterRate[k];

				//put links in both ways
				double r = agent_interaction_rate->GetNextRand() /* * i why multiply by i?*/ * weight_of_j / total_weight;
				r = std::min(1.0, r);	//cap at 1
				(*agents)[i]->encounterRate[j] = r;
				new_total_weight += r;

				r = agent_interaction_rate->GetNextRand() /* * i why multiply by i?*/ * weight_of_j / total_weight;
				r = std::min(1.0, r);	//cap at 1
				(*agents)[j]->encounterRate[i] = r;
				new_total_weight += r;
			}
			//add new weight from the new links
			total_weight += new_total_weight;
		}
	}

	void PrintConnections()
	{
		int num_agents = agents->size();
		for(int i = 0; i < num_agents; i++)
		{
			for(int j = 0; j < num_agents; j++)
				printf("%.3f ", (*agents)[i]->encounterRate[j]);
			printf("\n");
		}
	}

	std::vector<AgentType *> *agents;
};
