
template <class T>
int max_index(std::vector<T> &a)
{
	int max_index = 0;
	T max_value = a[0];

	for(int i = 1; i < a.size(); i++)
	{
		if(a[i] > max_value)
		{
			max_index = i;
			max_value = a[i];
		}
	}
	return max_index;
}

int round(double x)
{	return (int)(x + .5); }

class DiscreteFavors
{
public:
	DiscreteFavors()
	{
		
	}

	//pass in discount factor, c index, and c index of opponent, returns expected utility per unit time
	// assuming that the agent is deciding whether to perform the next favor
	double EvaluateExpectedTotalUtilityConstantStrategy(double df, int a, int other_a)
	{
		double total_utility = 0.0;
		total_utility -= C[a] / (1-sq(df));	//sum_t=0^infty df^(2*t) * C[a].  Includes every other turn starting at 0
		total_utility += W[other_a] * sq(df) / (1-sq(df));	//sum_t=1^infty df^(2*t) * W[a].  Includes every other turn starting at 0


		//multiply whole equation by (1-$df) to put in terms of unit time
		return 1/(1-df) * total_utility;
	}

	//takes in a discount factor and the maximum index that the opponent will offer and returns the maximum index it will offer
	int FindMaximumFavorSubjectToIndividualRationality(double df, int opponent_a)
	{
		for(int a = 0; a < C.size(); a++)
		{
			double value = EvaluateExpectedTotalUtilityConstantStrategy(df, a, opponent_a);
			if(value < 0)
				return a - 1;
		}
		return C.size()-1;
	}

	//pass in discount factor, opponent's discount factor, what belief other agent has of c, and reference to an array of c indexes (last one is repeated indefinitely)
	// returns the expected utility per unit time
	double EvaluateExpectedUtilityPerTimeDynamicStrategy(double df, double replacement_rate, double other_df, double other_a_belief, std::vector<int> &a)
	{
		//evaluate  utility at every timestep for which the agent has a strategy
		double total_u = 0;
		int max_time = 100; //a.size()
		for(int t = 0; t <= max_time; t++)
		{		
			//find out which strategy of the other player will provide it the most utility (given that it assumes this agent will be consistent)
			int other_a = FindMaximumFavorSubjectToIndividualRationality(other_df, round(other_a_belief));
			
			int action = (t < a.size() ? a[t] : a[a.size()-1]);
	 
			//figure out changes to utility contributed by only the next two interactions
			total_u -= pow(df, 2*t) * C[action];
			total_u += pow(df, 2*t + 1) * W[other_a];
			
			//opponent observes the action and uses that as its belief
			other_a_belief = (1-replacement_rate) * other_a_belief + replacement_rate * action;
		}
		
		//account for all future actions using the last action
		int other_a = FindMaximumFavorSubjectToIndividualRationality(other_df, round(other_a_belief));
		
		//figure out all future utility based on interactions and current knowledge
		total_u -= pow(df, 2*(max_time+1)) / (1 - sq(df)) * C[a[a.size()-1]];
		total_u -= pow(df, 2*(max_time+1)+1) / (1 - sq(df)) * W[other_a];
		
		//return the average expected return per game
		return (1-df) * total_u;
	}

	//pass in discount factor, opponent's discount factor, what belief other agent has of c, followed by the maximum length of a strategy
	// returns the best strategy
	void FindBestStrategyOfLengthL(double df, double replacement_rate, double other_df, int other_a_belief, int a_length, std::vector<int> &best_strategy)
	{
		//the number of possible values of C
		int num_values = C.size();
		
		//create the list of possible strategies
		std::vector< std::vector<int> > strategies;
		//for all strategy numbers
		int total_strategies = (int)pow((double)num_values, (double)(a_length));
		for(int strategy_num = 0; strategy_num < total_strategies; strategy_num++)
		{
			//take out digits of the number to create the strategy
			strategies.push_back(std::vector<int>());
			for(int i = 0; i < a_length; i++)
			{
				int base = (int)pow((double)num_values, (double)i);
				int value = (strategy_num / base) % num_values;
				strategies[strategy_num].push_back(value);
			}
			
		}
		
		//find best strategy
		std::vector<double> strategy_values;
		for(int i = 0; i < total_strategies; i++)
		{
			double value = EvaluateExpectedUtilityPerTimeDynamicStrategy(df, replacement_rate, other_df, other_a_belief, strategies[i]);
			strategy_values.push_back(value);
		}
		int mi = max_index(strategy_values);

		for(int i = 0; i < strategies[mi].size(); i++)
			best_strategy.push_back(strategies[mi][i]);
	}

	//returns a vector of vectors in the form of: player 1 payoff = payoffs[player 1 action][player 2 action]
	void ComputeExpectedPayoffsForAllStationaryStrategies(double df, std::vector< std::vector<double> > &payoffs)
	{
		for(int i = 0; i < C.size(); i++)
		{
			payoffs.push_back( std::vector<double>() );
			for(int j = 0; j < C.size(); j++)
				payoffs[i].push_back(EvaluateExpectedTotalUtilityConstantStrategy(df, i, j));
		}		
	}

	std::vector<int> W;
	std::vector<int> C;

};
