
//global of the favor game environment (used by agents)
extern FavorGameEnvironment *fge;

//const double min_c_check = -20;
//const double max_c_check = 15000;

const double min_c_check = -100;
const double max_c_check = 15000;

enum FAVOR_AGENT_RECIPROCITY_MODE
{
	FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL,
	FAVOR_AGENT_RECIPROCITY_DERIVATIVE_MATCH,
	FAVOR_AGENT_RECIPROCITY_SANCTION,
};

class MaxCGammaValueFavorEvaluatorBase : public Evaluator
{
public:
	double gamma, otherGamma;	//gamma of this agent and the other agent
	double c, otherC;			//c value of this agent and the other agent
	double w, otherW;
	double rAsk, rOffer;
	double replacementRateAsk, replacementRateOffer;
	//parameter position of each of the variables for this agent
	// if negative, then use the current value
	int gammaParamPos, otherGammaParamPos, cParamPos, otherCParamPos;
};


class MaxCGammaValueFavorEvaluatorRational : public MaxCGammaValueFavorEvaluatorBase
{
public:
	double Evaluate(std::vector<double> v)
	{
		if(gammaParamPos >= 0)	gamma = v[gammaParamPos];
		if(otherGammaParamPos >= 0)	otherGamma = v[otherGammaParamPos];
		if(cParamPos >= 0) c = v[cParamPos];
		if(otherCParamPos >= 0)	otherC = v[otherCParamPos];

		double discount = gamma / (1 - gamma);
		return c - discount * (
			rAsk * fge->W->GetExpectedValue() * fge->C->GetProbabilityInRange(fge->C->GetMinimumValue(), otherC)
			- rOffer * fge->C->GetExpectedValueInRange(fge->C->GetMinimumValue(), c));
	}
};

class MaxCGammaValueFavorEvaluatorDerivative : public MaxCGammaValueFavorEvaluatorBase
{
public:
	double Evaluate(std::vector<double> v)
	{
		if(gammaParamPos >= 0)	gamma = v[gammaParamPos];
		if(otherGammaParamPos >= 0)	otherGamma = v[otherGammaParamPos];
		if(cParamPos >= 0) c = v[cParamPos];
		if(otherCParamPos >= 0)	otherC = v[otherCParamPos];

		double og = otherGamma / (1 - otherGamma);
		return og * rOffer * fge->W->GetExpectedValue() * fge->C->GetPDFValue(c)
			- ( 1 + og * rAsk * otherC * fge->C->GetPDFValue(otherC) );
	}
};

class MaxCGammaValueFavorEvaluatorSanction : public MaxCGammaValueFavorEvaluatorBase
{
public:
	double Evaluate(std::vector<double> v)
	{
		if(gammaParamPos >= 0)	gamma = v[gammaParamPos];
		if(otherGammaParamPos >= 0)	otherGamma = v[otherGammaParamPos];
		if(cParamPos >= 0) c = v[cParamPos];
		if(otherCParamPos >= 0)	otherC = v[otherCParamPos];

		double discount = gamma / (1 - gamma);
		return w + discount * rOffer * fge->W->GetExpectedValue() * fge->C->GetProbabilityInRange(fge->C->GetMinimumValue(), otherC) - c;
	}
};

//takes in a game, the discount factor of the known agent, and whether the unknown discount factor to find is in the offer role
// returns the discount factor of a2 that is the boundary as to whether a2 would play P or R
// returns false if cannot compute a2's discount factor for whatever reason
template <class T>
bool ComputeGammaStar(FavorGame &fg, double known_discount_factor, bool find_offer_agent, double &unknown_discount_factor)
{
	//get the expected value of the other agent's discount factor
	T a1_solver, a2_solver;
	//a1, offering agent
	a1_solver.cParamPos = -1;				a1_solver.c = fg.c;
	a1_solver.otherCParamPos = 0;
	if(find_offer_agent)
	{
		a1_solver.gammaParamPos = 1;
		a1_solver.otherGammaParamPos = -1;	a1_solver.otherGamma = known_discount_factor;
	}
	else //find ask agent
	{
		a1_solver.gammaParamPos = -1;		a1_solver.gamma = known_discount_factor;
		a1_solver.otherGammaParamPos = 1;
	}

	a1_solver.w = 0;	//giving favor to other agent
	a1_solver.otherW = fg.w;

	a1_solver.rAsk = fg.a2->encounterRate[fg.a1->agentNumber];
	a1_solver.rOffer = fg.a1->encounterRate[fg.a2->agentNumber];

	a1_solver.replacementRateAsk = fg.a2->replacementRate;
	a1_solver.replacementRateOffer = fg.a1->replacementRate;

	//a2, asking agent
	a2_solver.cParamPos = 0;
	a2_solver.otherCParamPos = -1;		a2_solver.otherC = fg.c;
	if(find_offer_agent)
	{
		a2_solver.gammaParamPos = -1;		a2_solver.gamma = known_discount_factor;
		a2_solver.otherGammaParamPos = 1;			
	}
	else //find ask agent
	{
		a2_solver.gammaParamPos = 1;
		a2_solver.otherGammaParamPos = -1;	a2_solver.otherGamma = known_discount_factor;
	}

	a2_solver.w = fg.w;	//receiving favor
	a2_solver.otherW = 0;

	a2_solver.rAsk = fg.a1->encounterRate[fg.a2->agentNumber];
	a2_solver.rOffer = fg.a2->encounterRate[fg.a1->agentNumber];

	a2_solver.replacementRateAsk = fg.a1->replacementRate;
	a2_solver.replacementRateOffer = fg.a2->replacementRate;

	//parameters
	std::vector<Evaluator *> tests;
	tests.push_back(&a1_solver);
	tests.push_back(&a2_solver);

	//param pos 0 c_a1, param pos 1 is gamma_a2
	// note that gamma_a2 is undefined for 1, so need to go just under 1
	std::vector<double> min;		min.push_back(min_c_check);		min.push_back(0);
	std::vector<double> max;		max.push_back(max_c_check);		max.push_back(1.0 - 0.0001);
	std::vector<double> tolerance;	tolerance.push_back(0.1);		tolerance.push_back(.005);

	std::vector<double> results;	results.push_back(0);							results.push_back(0);

	if(!SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
		return false;

	unknown_discount_factor = results[1];
	return true;
}

//takes in a game and the discount factors of each agent, returns the c values for the two agents
// returns false if cannot compute the c values for whatever reason
template <class T>
bool ComputeCStar(FavorGame &fg, double a1_discount_factor, double a2_discount_factor,
					double &c_a1, double &c_a2)
{
	//get the expected value of the other agent's discount factor
	T a1_solver, a2_solver;
	//a1, offering agent
	a1_solver.cParamPos = 0;
	a1_solver.otherCParamPos = 1;
	a1_solver.gammaParamPos = -1;		a1_solver.gamma = a1_discount_factor;
	a1_solver.otherGammaParamPos = -1;	a1_solver.otherGamma = a2_discount_factor;

	a1_solver.w = 0;	//giving favor to other agent
	a1_solver.otherW = fg.w;

	a1_solver.rAsk = fg.a2->encounterRate[fg.a1->agentNumber];
	a1_solver.rOffer = fg.a1->encounterRate[fg.a2->agentNumber];

	a1_solver.replacementRateAsk = fg.a2->replacementRate;
	a1_solver.replacementRateOffer = fg.a1->replacementRate;

	//a2, asking agent
	a2_solver.cParamPos = 1;
	a2_solver.otherCParamPos = 0;
	a2_solver.gammaParamPos = -1;		a2_solver.gamma = a2_discount_factor;
	a2_solver.otherGammaParamPos = -1;	a2_solver.otherGamma = a1_discount_factor;

	a2_solver.w = fg.w;	//receiving favor
	a2_solver.otherW = 0;

	a2_solver.rAsk = fg.a1->encounterRate[fg.a2->agentNumber];
	a2_solver.rOffer = fg.a2->encounterRate[fg.a1->agentNumber];

	a1_solver.replacementRateAsk = fg.a1->replacementRate;
	a1_solver.replacementRateOffer = fg.a2->replacementRate;

	//parameters
	std::vector<Evaluator *> tests;
	tests.push_back(&a1_solver);
	tests.push_back(&a2_solver);

	//param pos 0 c_a_1, param pos 1 is c_a_2
	std::vector<double> min;		min.push_back(min_c_check);		min.push_back(min_c_check);
	std::vector<double> max;		max.push_back(max_c_check);		max.push_back(max_c_check);
	std::vector<double> tolerance;	tolerance.push_back(0.1);		tolerance.push_back(0.1);
	std::vector<double> results;	results.push_back(0);							results.push_back(0);

	if(!SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
		return false;

	c_a1 = results[0];
	c_a2 = results[1];

/*
//uncomment to find the range of c values to use
static double c_min = INFINITY;
static double c_max = -INFINITY;
c_min = std::min(c_min, c_a1);
c_min = std::min(c_min, c_a2);
c_max = std::max(c_max, c_a1);
c_max = std::max(c_max, c_a2);
printf("%f, %f\n", c_min, c_max);
//*/
	return true;
}


class ReciprocatingFavorAgent : public FavorAgent
{
public:
	ReciprocatingFavorAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate,
								bool perfect_signalling, FAVOR_AGENT_RECIPROCITY_MODE reciprocity_mode)
		: FavorAgent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{
		for(int i = 0; i < num_agents; i++)
		{
			observationHistory.push_back(ObservationHistory());
			observationSelfHistory.push_back(ObservationHistory());
		}

		perfectSignalling = perfect_signalling;
		reciprocityMode = reciprocity_mode;
	}

	bool ComputeGammaStar(FavorGame &fg, double known_discount_factor, bool find_offer_agent, double &unknown_discount_factor)
	{
		
		if(!::ComputeGammaStar<MaxCGammaValueFavorEvaluatorRational>(fg, known_discount_factor, find_offer_agent, unknown_discount_factor))
			return false;

		switch(reciprocityMode)
		{
		case FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL:
			return true;
		case FAVOR_AGENT_RECIPROCITY_DERIVATIVE_MATCH:
			{
				double new_unknown_discount_factor;
				if(!::ComputeGammaStar<MaxCGammaValueFavorEvaluatorDerivative>(fg, known_discount_factor, find_offer_agent, new_unknown_discount_factor))
					return false;
				unknown_discount_factor = std::max(new_unknown_discount_factor, unknown_discount_factor);
				return true;
			}
		case FAVOR_AGENT_RECIPROCITY_SANCTION:
			{
				double new_unknown_discount_factor;
				if(!::ComputeGammaStar<MaxCGammaValueFavorEvaluatorSanction>(fg, known_discount_factor, find_offer_agent, new_unknown_discount_factor))
				return false;
				unknown_discount_factor = std::max(new_unknown_discount_factor, unknown_discount_factor);
				return true;
			}
		}
		return false;
	}

	bool ComputeCStar(FavorGame &fg, double a1_discount_factor, double a2_discount_factor, double &c_a1, double &c_a2)
	{
		if(!::ComputeCStar<MaxCGammaValueFavorEvaluatorRational>(fg, a1_discount_factor, a2_discount_factor, c_a1, c_a2))
			return false;
		switch(reciprocityMode)
		{
		case FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL:
			return true;
		case FAVOR_AGENT_RECIPROCITY_DERIVATIVE_MATCH:
			{
				double new_c_a1, new_c_a2;
				if(!::ComputeCStar<MaxCGammaValueFavorEvaluatorDerivative>(fg, a1_discount_factor, a2_discount_factor, new_c_a1, new_c_a2))
					return false;
				c_a1 = std::min(c_a1, new_c_a1);
				c_a2 = std::min(c_a2, new_c_a2);
				return true;
			}
		case FAVOR_AGENT_RECIPROCITY_SANCTION:
			{
				double new_c_a1, new_c_a2;
				if(!::ComputeCStar<MaxCGammaValueFavorEvaluatorSanction>(fg, a1_discount_factor, a2_discount_factor, new_c_a1, new_c_a2))
					return false;
				c_a1 = std::min(c_a1, new_c_a1);
				c_a2 = std::min(c_a2, new_c_a2);
				return true;
			}
		}
		return false;
	}

	//returns the total utility an agent can expect from future plays against all agents listed in the vector
	double ComputeContinuationValue(double discount_factor, std::vector<int> &agents, 
									std::vector<ObservationHistory> &observation_history,
									std::vector<ObservationHistory> &observation_self_history,
									double cur_time)
	{
		double total_utility = 0.0;

		FavorGame fg;
		for(std::vector<int>::iterator i = agents.begin(); i != agents.end(); i++)
		{
			fg.a1 = this;
			fg.a2 = (*fge->agents)[*i];
			double c_a1, c_a2;
			if(ComputeCStar(fg, observation_self_history[*i].GetExpectedValue(1 - fg.a1->replacementRate, cur_time),
				observation_history[*i].GetExpectedValue(1 - fg.a2->replacementRate, cur_time), c_a1, c_a2))
			{
				//utility from ask role
				total_utility += fg.a2->encounterRate[agentNumber]
								* fge->C->GetProbabilityInRange(0, c_a2) * fge->W->GetExpectedValue();

				//utility from offer role
				total_utility -= encounterRate[fg.a2->agentNumber]
								* fge->C->GetProbabilityInRange(0, c_a1) * fge->C->GetExpectedValueInRange(0, c_a1);
			}
		}

		return discount_factor / (1 - discount_factor) * total_utility;
	}

	//returns the optimal action for an offer agent
	// if answering_question is true, then it may return FAV_ACTION_N
	FavorAgentAction GetOptimalActionForOfferAgent(FavorGame &fg, double discount_factor_offer_agent, double discount_factor_ask_agent, bool answering_question = false)
	{
		double c_a1, c_a2;
		if(!ComputeCStar(fg, discount_factor_offer_agent, discount_factor_ask_agent, c_a1, c_a2))
		{
			if(answering_question)
				return FAV_ACTION_N;
			return FAV_ACTION_R;
		}

		if(fg.c < c_a1)
			return FAV_ACTION_P;
		return FAV_ACTION_R;
	}

	double discountFactorBeforeObservation;

	virtual FavorAgentAction ChooseAction(FavorGame &fg)
	{
		return ChooseActionFromExpectedUtilityChanges(fg);
		//return NaiivelyChooseAction(fg);
	}

	FavorAgentAction ChooseActionFromExpectedUtilityChanges(FavorGame &fg)
	{
		if(fg.a2 == this)
		{
			discountFactorBeforeObservation = observationHistory[fg.a1->agentNumber].GetExpectedValue(1 - fg.a1->replacementRate, fg.time);
			return FAV_ACTION_P;
		}
		discountFactorBeforeObservation = 0;

		int other_agent_num = fg.a2->agentNumber;
		double a2_discount_factor_estimate = observationHistory[fg.a2->agentNumber].GetExpectedValue(1 - fg.a2->replacementRate, fg.time);

		double a1_discount_factor_observation;
		//see what the observation will be
		if(!ComputeGammaStar(fg, a2_discount_factor_estimate, true, a1_discount_factor_observation))
			return GetOptimalActionForOfferAgent(fg, discountFactor, a2_discount_factor_estimate);

		//compute utilities of playing different actions based on how it will affect other agents' perceptions
		std::vector<int> agents;	agents.push_back(other_agent_num);

		//compute utility given playing P
		std::vector<ObservationHistory> observation_self_history;
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(a1_discount_factor_observation, 1, fg.time);
		observation_self_history[other_agent_num].HandleInconsistantObservations(1 - fg.a1->replacementRate, fg.time, perfectSignalling);

		double V_P = ComputeContinuationValue(discountFactor, agents, observationHistory, observation_self_history, fg.time+1);
		V_P -= fg.c;

		//compute utility given playing R
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(0, a1_discount_factor_observation, fg.time);
		observation_self_history[other_agent_num].HandleInconsistantObservations(1 - fg.a1->replacementRate, fg.time, perfectSignalling);

		double V_R = ComputeContinuationValue(discountFactor, agents, observationHistory, observation_self_history, fg.time+1);

		if(V_R > V_P)
			return FAV_ACTION_R;
		return FAV_ACTION_P;
	}

	FavorAgentAction NaiivelyChooseAction(FavorGame &fg)
	{
		if(fg.a2 == this)
		{
			discountFactorBeforeObservation = observationHistory[fg.a1->agentNumber].GetExpectedValue(1 - fg.a1->replacementRate, fg.time);
			return FAV_ACTION_P;
		}
		discountFactorBeforeObservation = 0.0;

		double expected_a2_discount_factor = observationHistory[fg.a2->agentNumber].GetExpectedValue(1 - fg.a2->replacementRate, fg.time);
		return GetOptimalActionForOfferAgent(fg, discountFactor, expected_a2_discount_factor);
	}

	void RecordObservation(ObservationHistory &oh, double oh_replacement_rate, FavorGame &fg, double known_discount_factor, bool find_offer_agent)
	{
		double unknown_discount_factor;
		if(ComputeGammaStar(fg, known_discount_factor, find_offer_agent, unknown_discount_factor))
		{
			if(fg.a1Action == FAV_ACTION_P)
				oh.AddObservation(unknown_discount_factor, 1.0, fg.time);
			else if(fg.a1Action == FAV_ACTION_R)
				oh.AddObservation(0.0, unknown_discount_factor, fg.time);
			oh.HandleInconsistantObservations(1 - oh_replacement_rate, fg.time, perfectSignalling);
		}
	}

	virtual void MakeObservation(FavorGame &fg)
	{
		//want to determine the discount factor boundary that this game represents

		//add observation
		if(fg.a1 == this)	//offer agent
		{
			//record observation that opponent (in the ask role (a2)) will make about this agent (a1)
			double a2_discount_factor_estimate = observationHistory[fg.a2->agentNumber].GetExpectedValue(1 - fg.a2->replacementRate, fg.time);
			RecordObservation(observationSelfHistory[fg.a2->agentNumber], fg.a1->replacementRate, fg, a2_discount_factor_estimate, true);
		}
		else if(fg.a2 == this)	//ask agent
		{	
			//record observation about opponent in the offer role (a1)
			//compute discount factor as other agent is seeing it
			double a2_discount_factor_estimate = observationSelfHistory[fg.a1->agentNumber].GetExpectedValue(1 - replacementRate, fg.time);
			RecordObservation(observationHistory[fg.a1->agentNumber], fg.a1->replacementRate, fg, a2_discount_factor_estimate, true);
		}
	}
	
	virtual void PrintInfo()
	{
		printf("agent %d:\t$: %f\tdf: %f\tr: %f\n", agentNumber, totalUtility, discountFactor, replacementRate);
	}

	virtual void GetStringOnOpponent(char *buffer, FavorAgent *fa, double cur_time)
	{
		if(discountFactorBeforeObservation == 0.0)
			sprintf(buffer, "%.3f", observationHistory[fa->agentNumber].GetExpectedValue(1 - fa->replacementRate, cur_time));
		else
			sprintf(buffer, "%.3f->%.3f", discountFactorBeforeObservation, observationHistory[fa->agentNumber].GetExpectedValue(1 - fa->replacementRate, cur_time));
	}

	//this agent's direct observations of other agents
	std::vector<ObservationHistory> observationHistory;
	//other agents' direct observations of this agent
	std::vector<ObservationHistory> observationSelfHistory;

	//true if the agent believes all signals are perfect and contain no noise, false if it believes randomness occurs
	bool perfectSignalling;

	//specifies which reciprocity mode for the agent to use
	FAVOR_AGENT_RECIPROCITY_MODE reciprocityMode;
};

class CommunicatingReciprocatingFavorAgent : public ReciprocatingFavorAgent
{
public:
	CommunicatingReciprocatingFavorAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate, int min_num_observations_to_judge, bool perfect_signalling, FAVOR_AGENT_RECIPROCITY_MODE reciprocity_mode)
		: ReciprocatingFavorAgent(agent_number, num_agents, discount_factor_distribution, replacement_rate, perfect_signalling, reciprocity_mode)
	{

	minNumObservationsToJudge = min_num_observations_to_judge;
/*OLD CODE...DELETE THIS?
// say a1 is deciding whether to offer a2 a favor, and asks a3 if a3 would offer a favor in the current game to a2
	//  questions[a3][a2] would contain a vector of the games of which a1 asked a3 a "question" about a2
	// the game's a1Action field contains the "answer"
	std::vector<std::vector<std::vector<FavorGame>>> questions;

		for(int i = 0; i < num_agents; i++)
		{
			questions.push_back(std::vector<std::vector<FavorGame>>());
			for(int j = 0; j < num_agents; j++)
				questions[i].push_back(std::vector<FavorGame>());
		}
*/
	}

	virtual FavorAgentAction ChooseAction(FavorGame &fg)
	{
		return ChooseActionFromCommunication(fg);
		//return ChooseActionFromExpectedUtilityChanges(fg);
		//return NaiivelyChooseAction(fg);
	}

	FavorAgentAction ChooseActionFromCommunication(FavorGame &fg)
	{
		if(fg.a2 == this)
		{
			discountFactorBeforeObservation = observationHistory[fg.a1->agentNumber].GetExpectedValue(1 - fg.a1->replacementRate, fg.time);
			return FAV_ACTION_P;
		}
		discountFactorBeforeObservation = 0.0;

		//ask questions of all of the other agents
		for(int i = 0; i < fge->agents->size(); i++)
		{
			//don't ask self or the agent playing against
			if(i == agentNumber || i == fg.a2->agentNumber)
				continue;

			//don't ask an agent that don't know much about
			if(observationHistory[i].observations.size() < minNumObservationsToJudge)
				continue;

			//make game with other agent in it, ask its response
			FavorGame q = fg;
			q.a1 = (*fge->agents)[i];
			FavorAgentAction faa = ((CommunicatingReciprocatingFavorAgent *)q.a1)->ChooseActionFromPreviousCommunication(q, true);

			//if the agent doesn't say no, then record the game
			if(faa != FAV_ACTION_N)
			{
				q.a1Action = faa;
				questions.push_back(q);
			}
		}

		//compute action from everything other than new communications
		return ChooseActionFromPreviousCommunication(fg, false);
	}

	//if answering_question is true, then can return FAV_ACTION_N when don't know
	FavorAgentAction ChooseActionFromPreviousCommunication(FavorGame &fg, bool answering_question)
	{
		if(fg.a2 == this)
		{
			discountFactorBeforeObservation = observationHistory[fg.a1->agentNumber].GetExpectedValue(1 - fg.a1->replacementRate, fg.time);
			return FAV_ACTION_P;
		}
		discountFactorBeforeObservation = 0;

		//only answer if have more than 2 observations
		if(answering_question && observationHistory[fg.a2->agentNumber].observations.size() < minNumObservationsToJudge)
			return FAV_ACTION_N;

		//create temporary history of this agent based communication
		std::vector<ObservationHistory> observation_history;
		DeepCopyObservationHistory(observationHistory, observation_history);

		//work through agent communications in the order they were received
		for(std::vector<FavorGame>::iterator g = questions.begin(); g != questions.end(); g++)
		{
			//should only judge others when have enough information for both agents in the game in order to judge
			if(observation_history[g->a1->agentNumber].observations.size() < minNumObservationsToJudge
					|| observation_history[g->a2->agentNumber].observations.size() < minNumObservationsToJudge)
				continue;

			//compute discount factor as other agent is seeing it
			double a1_discount_factor_estimate = observation_history[g->a1->agentNumber].GetExpectedValue(1 - g->a1->replacementRate, fg.time);
			double a2_discount_factor_estimate = observation_history[g->a2->agentNumber].GetExpectedValue(1 - g->a2->replacementRate, fg.time);

			//a1Action contains the answer given in the hypothetical game between a1 and a2
			//see if a1 would still play the same answer against a2, given current knowledge
			FavorAgentAction old_a1_action = g->a1Action;
			FavorAgentAction new_a1_action = GetOptimalActionForOfferAgent(*g, a1_discount_factor_estimate, a2_discount_factor_estimate, true);
			//if don't know, don't hold it against the agent
			if(new_a1_action == FAV_ACTION_N)
				continue;

			//if so, make normal observations for both a1 and a2
			if(new_a1_action == old_a1_action)
			{
				//record observation about opponent in the offer role (a1)
				RecordObservation(observation_history[g->a1->agentNumber], g->a1->replacementRate, *g, a2_discount_factor_estimate, true);

				//record observation about opponent in the ask role (a2)
				RecordObservation(observation_history[g->a2->agentNumber], g->a2->replacementRate, *g, a1_discount_factor_estimate, false);
			}
			else //a1 lied about whether it would play the action it did - only record observations about a1
			{
				//if recommended playing R, then assume discount factor is below that required for the game
				if(old_a1_action == FAV_ACTION_R)
				{
					double a1_discount_factor_comm_obs;
					if(ComputeGammaStar(*g, a2_discount_factor_estimate, true, a1_discount_factor_comm_obs))
					{
						observation_history[g->a1->agentNumber].AddObservation(0.0, a1_discount_factor_comm_obs, g->time);
						observation_history[g->a1->agentNumber].HandleInconsistantObservations(1 - g->a1->replacementRate, g->time, perfectSignalling);
					}					
				}
				else //if recommended playing P, then make a "really low" observation
				{
					observation_history[g->a1->agentNumber].AddObservation(0.0, .05, g->time);
					observation_history[g->a1->agentNumber].HandleInconsistantObservations(1 - g->a1->replacementRate, g->time, perfectSignalling);
				}

			}
		}

		int other_agent_num = fg.a2->agentNumber;
		double a2_discount_factor_estimate = observation_history[fg.a2->agentNumber].GetExpectedValue(1 - fg.a2->replacementRate, fg.time);

		double a1_discount_factor_observation;
		//see what the observation will be
		if(!ComputeGammaStar(fg, a2_discount_factor_estimate, true, a1_discount_factor_observation))
			return GetOptimalActionForOfferAgent(fg, discountFactor, a2_discount_factor_estimate, answering_question);

		//compute utilities of playing different actions based on how it will affect other agents' perceptions
		std::vector<int> agents;	agents.push_back(other_agent_num);

		//compute utility given playing P
		std::vector<ObservationHistory> observation_self_history;
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);
		double current_a1_expected_discount_factor = observationSelfHistory[other_agent_num].GetExpectedValue(1 - fg.a1->replacementRate, fg.time);

		observation_self_history[other_agent_num].AddObservation(a1_discount_factor_observation, 1, fg.time);
		observation_self_history[other_agent_num].HandleInconsistantObservations(1 - fg.a1->replacementRate, fg.time, perfectSignalling);

		double V_P = ComputeContinuationValue(discountFactor, agents, observation_history, observation_self_history, fg.time+1);
		if(V_P == 0.0 && answering_question)
			return FAV_ACTION_N;
		V_P -= fg.c;

		//compute utility given playing R
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(0, a1_discount_factor_observation, fg.time);
		observation_self_history[other_agent_num].HandleInconsistantObservations(1 - fg.a1->replacementRate, fg.time, perfectSignalling);

		double V_R = ComputeContinuationValue(discountFactor, agents, observation_history, observation_self_history, fg.time+1);
		if(V_R == 0.0 && answering_question)
			return FAV_ACTION_N;

		if(V_R > V_P)
			return FAV_ACTION_R;
		return FAV_ACTION_P;
	}

	//minimum number of observations that another agent claims before it will consider the observations valid
	int minNumObservationsToJudge;

	//stores a list of games for each pair of agents
	// say a1 is deciding whether to offer a2 a favor, and asks a3 if a3 would offer a favor in the current game to a2
	//  questions[a3][a2] would contain a vector of the games of which a1 asked a3 a "question" about a2
	// the game's a1Action field contains the "answer"
	std::vector<FavorGame> questions;

};
