//global of the favor game environment (used by agents)
extern FavorGameEnvironment *fge;

class FavorAgentProbabilisticReciprocityAgent : public FavorAgent
{
public:
	FavorAgentProbabilisticReciprocityAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate)
		: FavorAgent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{
		agentBallance.resize(num_agents, 0);

	}

	virtual FavorAgentAction ChooseAction(FavorGame &fg)
	{
		if(fg.a2 == this)
			return FAV_ACTION_P;
		//return true;

		double tau = 0.75;
		double beta = 0.5;
		double average_task_cost = fge->C->GetExpectedValue();
		double current_cost = fg.c;
		double ballance = agentBallance[fg.a2->agentNumber];

		double pr = 1.0 / (1.0 + exp(
									(current_cost - beta * average_task_cost - ballance) / tau ));

		double r = genrand_real1();
		if(r < pr)
			return FAV_ACTION_P;
		return FAV_ACTION_R;
	}

	virtual void MakeObservation(FavorGame &fg)
	{	
		if(fg.a2 == this)
		{
			if(fg.a1Action == FAV_ACTION_P)
			{
				int other_agent_num = fg.a1->agentNumber;
				agentBallance[other_agent_num] += fg.w;
			}

		}
		else //was askee (a1)
		{
			if(fg.a1Action == FAV_ACTION_P)
			{
				int other_agent_num = fg.a2->agentNumber;
				agentBallance[other_agent_num] -= fg.c;
			}

		}
	}

	//the index is the agent number that has the ballance
	std::vector<double> agentBallance;
};


class FixedFavorAgent : public FavorAgent
{
public:
	FixedFavorAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate, bool optimistic)
		: FavorAgent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{
		isOptimistic = optimistic;
	}

	virtual FavorAgentAction ChooseAction(FavorGame &fg)
	{
		if(isOptimistic)
			return FAV_ACTION_P;
		else
			return FAV_ACTION_R;
	}

	virtual void MakeObservation(FavorGame &fg)
	{	
	}

	bool isOptimistic;
};