enum FavorAgentAction
{
	FAV_ACTION_P,		//positive
	FAV_ACTION_R,		//refrain
	FAV_ACTION_N,		//no answer (for communication: no comment)
};

class FavorAgent;

//used to describe a single favor game to be passed into an agent
// a2 asks a1
// a2 always asks, a1 decides whether to offer
class FavorGame
{
public:
	FavorAgent *a1, *a2;
	FavorAgentAction a1Action, a2Action;
	double c, w;
	double time;
};

class FavorAgent : public Agent
{
public:
	FavorAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate)
		: Agent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{

	}
	
	//if agent is a1, then is selling, if a2, then buying
	virtual FavorAgentAction ChooseAction(FavorGame &fg)
	{
		double r = genrand_real1();
		if(r < 0.5)
			return FAV_ACTION_R;
		else
			return FAV_ACTION_P;
	}

	virtual void MakeObservation(FavorGame &fg)
	{	}

	virtual void GetStringOnOpponent(char *buffer, FavorAgent *fa, double cur_time)
	{	buffer[0] = '\0';	}

protected:

};

class FavorGameEnvironment : public MultiAgentSimulation<FavorAgent, FavorGame>
{
public:
	FavorGameEnvironment(RandomVariable *c, RandomVariable *w, std::vector<FavorAgent *> *agent_list)
	{
		C = c; W = w;
		agents = agent_list;
	}
	
	void AssignPayoffs(FavorGame &fg)
	{
		//no information gained if either played R
		if(fg.a1Action == FAV_ACTION_R || fg.a2Action == FAV_ACTION_R)
			return;

		fg.a1->totalUtility -= fg.c;
		fg.a2->totalUtility += fg.w;
	}
	
	void PrintGameInfo(FavorGame &fg)
	{
		char a1, a2;
		switch(fg.a1Action)
		{
			case FAV_ACTION_R: a1 = 'R';	break;
			case FAV_ACTION_P: a1 = 'P';	break;
			case FAV_ACTION_N: a1 = 'N';	break;
		}
		switch(fg.a2Action)
		{
			case FAV_ACTION_R: a2 = 'R';	break;
			case FAV_ACTION_P: a2 = 'P';	break;
			case FAV_ACTION_N: a2 = 'N';	break;
		}
		char buffer_a1[128], buffer_a2[128];
		fg.a1->GetStringOnOpponent(buffer_a1, fg.a2, fg.time);
		fg.a2->GetStringOnOpponent(buffer_a2, fg.a1, fg.time);
		printf(" a1:%d:%s:%c a2:%d:%s:%c c:%f w:%f\n",	fg.a1->agentNumber, buffer_a1, a1,
														fg.a2->agentNumber, buffer_a2, a2,
														fg.c, fg.w);
	}
	
	bool ExecuteRound(double cur_time)
	{
		//check for discount factor replacements
//		for(int i = 0; i < agents->size(); i++)
//			(*agents)[i]->UpdateDiscountFactor(cur_time);

		//for every combination of agents as buyer and seller, see if should play a game
		for(int i = 0; i < agents->size(); i++)
			for(int j = 0; j < agents->size(); j++)
			{
				//an agent can't play itself
				if(i == j)
					continue;

				if(genrand_real1() >= (*agents)[i]->encounterRate[j])
					continue;

				//build a game to play
				FavorGame fg;
				fg.a1 = (*agents)[i];
				fg.a2 = (*agents)[j];

				fg.c = C->GetNextRand();
				fg.w = W->GetNextRand();

				fg.time = cur_time;

				//get agents' moves
				fg.a1Action = fg.a1->ChooseAction(fg);
				fg.a2Action = fg.a2->ChooseAction(fg);

				//store results and payoffs
				fg.a1->MakeObservation(fg);
				fg.a2->MakeObservation(fg);

				AssignPayoffs(fg);

				PrintGameInfo(fg);
			}
		return false;	//the infinite game is never done
	}

	RandomVariable *C, *W;
};
