
//global of the market game environment (used by agents)
extern MarketGameEnvironment *mge;

class MaxCMGammaValueMarketEvaluator : public Evaluator
{
public:
	double gamma, otherGamma;	//gamma of this agent and the other agent
	double c, m;			//c/m value of this agent and the other agent
	bool seller;			//set to true if seller
	double rSell, rBuy; //ask offer
	//parameter position of each of the variables for this agent
	// if negative, then use the current value
	int gammaParamPos, otherGammaParamPos, cParamPos, mParamPos;

	double Evaluate(std::vector<double> v)
	{
		if(gammaParamPos >= 0)	gamma = v[gammaParamPos];
		if(otherGammaParamPos >= 0)	otherGamma = v[otherGammaParamPos];
		if(cParamPos >= 0) c = v[cParamPos];
		if(mParamPos >= 0) m = v[mParamPos];

		if(seller)
		{
			return gamma - c / (c + rSell * mge->C->GetProbabilityInRange(c, INFINITY)
											* mge->M->GetProbabilityInRange(0, m)
											* (mge->M->GetExpectedValueInRange(0, m)
												- mge->C->GetExpectedValueInRange(c, INFINITY) ));
		}
		else //buyer
		{
			return gamma - m / (m + rBuy * mge->M->GetProbabilityInRange(m, INFINITY)
											* mge->C->GetProbabilityInRange(0, c)											
											* (mge->W->GetExpectedValue()
												- mge->M->GetExpectedValueInRange(m, INFINITY) ));
		}
	}
};


class ADFDAgent : public MarketAgent
{
public:
	ADFDAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate, int min_num_observations_to_judge)
		: MarketAgent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{
		for(int i = 0; i < num_agents; i++)
		{
			observationHistory.push_back(ObservationHistory());
			observationSelfHistory.push_back(ObservationHistory());
		}

		minNumObservationsToJudge = min_num_observations_to_judge;
	}
/*	
	MarketGameAction ChooseAction(MarketGame &mg)
	{
		return MKT_ACTION_R;
	}
*/

	//takes in a game and returns the corresponding gamma values
	static void ComputeGammaStar(MarketGame &mg, double &seller_discount_factor, double &buyer_discount_factor)
	{
		double r = mg.a1->encounterRate[mg.a2->agentNumber];
		double c = mg.c, m = mg.m;
		seller_discount_factor = c / (c + r * mge->C->GetProbabilityInRange(c, INFINITY)
											* mge->M->GetProbabilityInRange(0, m)
											* (mge->M->GetExpectedValueInRange(0, m)
												- mge->C->GetExpectedValueInRange(c, INFINITY) ));
		buyer_discount_factor =	m / (m + r * mge->M->GetProbabilityInRange(m, INFINITY)
											* mge->C->GetProbabilityInRange(0, c)
											* (mge->W->GetExpectedValue()
												- mge->M->GetExpectedValueInRange(m, INFINITY) ));
	}

	//takes in a game and the discount factors of each agent, returns the c and m values for the two agents
	// returns false if cannot compute the c and m values for whatever reason
	static bool ComputeCMStar(MarketGame &mg, double a1_discount_factor, double a2_discount_factor,
									double &c_a1, double &m_a2)
	{
		//get the expected value of the other agent's discount factor
		MaxCMGammaValueMarketEvaluator a1_solver, a2_solver;
		a1_solver.seller = true;
		a1_solver.cParamPos = 0;
		a1_solver.mParamPos = 1;
		a1_solver.gammaParamPos = -1;		a1_solver.gamma = a1_discount_factor;
		a1_solver.otherGammaParamPos = -1;	a1_solver.otherGamma = a2_discount_factor;

		a1_solver.rBuy = mg.a2->encounterRate[mg.a1->agentNumber];
		a1_solver.rSell = mg.a1->encounterRate[mg.a2->agentNumber];

		a2_solver.seller = false;
		a2_solver.cParamPos = 0;
		a2_solver.mParamPos = 1;
		a2_solver.gammaParamPos = -1;		a2_solver.gamma = a2_discount_factor;
		a2_solver.otherGammaParamPos = -1;	a2_solver.otherGamma = a1_discount_factor;

		a2_solver.rBuy = mg.a1->encounterRate[mg.a2->agentNumber];
		a2_solver.rSell = mg.a2->encounterRate[mg.a1->agentNumber];

		//parameters
		std::vector<Evaluator *> tests;
		tests.push_back(&a1_solver);
		tests.push_back(&a2_solver);

		//param pos 0 c_a_1, param pos 1 is m_a_2
		std::vector<double> min;		min.push_back(mge->C->GetMinimumValue()+.05);		min.push_back(mge->M->GetMinimumValue()+.05);
		std::vector<double> max;		max.push_back(mge->C->GetMaximumValue()-.05);		max.push_back(mge->M->GetMaximumValue()-.05);
		std::vector<double> tolerance;	tolerance.push_back(0.1);							tolerance.push_back(0.1);
		std::vector<double> results;	results.push_back(0);							results.push_back(0);

		if(!SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
			return false;

		c_a1 = results[0];
		m_a2 = results[1];
		return true;
	}

	//returns the total utility an agent can expect from future plays against all agents listed in the vector
	double ComputeContinuationValue(double discount_factor, std::vector<int> &agents, 
									std::vector<ObservationHistory> &observation_history,
									std::vector<ObservationHistory> &observation_self_history,
									double cur_time)
	{
		double total_utility = 0.0;

		MarketGame mg;
		for(std::vector<int>::iterator i = agents.begin(); i != agents.end(); i++)
		{
			double own_discount_factor = observation_self_history[*i].GetExpectedValue(1 - replacementRate, cur_time);
			double other_discount_factor = observation_history[*i].GetExpectedValue(1 - (*mge->agents)[*i]->replacementRate, cur_time);

			double c, m;

			mg.a1 = this;
			mg.a2 = (*mge->agents)[*i];
			if(ComputeCMStar(mg, own_discount_factor, other_discount_factor, c, m))
			{
				//utility from buy role
				total_utility += mg.a2->encounterRate[agentNumber]
								* mge->C->GetProbabilityInRange(0, c)
								* mge->M->GetProbabilityInRange(0, m)
								* (mge->M->GetExpectedValueInRange(0, m)
									- mge->C->GetExpectedValueInRange(0, c) );
			}

			mg.a2 = this;
			mg.a1 = (*mge->agents)[*i];
			if(ComputeCMStar(mg, other_discount_factor, own_discount_factor, c, m))
			{
				//utility from sell role
				total_utility += encounterRate[mg.a2->agentNumber]
								* mge->M->GetProbabilityInRange(0, m)
								* mge->C->GetProbabilityInRange(0, c)
								* (mge->W->GetExpectedValue()
									- mge->M->GetExpectedValueInRange(0, m) );
			}
		}

		return discount_factor / (1 - discount_factor) * total_utility;
	}


	//returns the safe action for the specified role
	MarketAgentAction GetSafeAction(MarketGame &mg, double discount_factor_sell_agent, double discount_factor_buy_agent)
	{
		double seller_discount_factor_observation, buyer_discount_factor_observation;
		ComputeGammaStar(mg, seller_discount_factor_observation, buyer_discount_factor_observation);

		if(discount_factor_sell_agent >= seller_discount_factor_observation
				&& discount_factor_buy_agent >= buyer_discount_factor_observation)
			return MKT_ACTION_P;
		return MKT_ACTION_R;

		/*
		double c, m;
		if(!ComputeCMStar(mg, discount_factor_sell_agent, discount_factor_buy_agent, c, m))
			return MKT_ACTION_R;

		if(mg.c < c && mg.m < m)
			return MKT_ACTION_P;
		return MKT_ACTION_R;
		*/
	}

	virtual MarketAgentAction ChooseAction(MarketGame &mg)
	{
		return ChooseActionFromCommunication(mg);
		//return ChooseActionFromExpectedUtilityChanges(mg);
		//return NaiivelyChooseAction(mg);
		//return MKT_ACTION_P;
	}

	MarketAgentAction ChooseActionFromCommunication(MarketGame &mg)
	{
		//find other agent's number
		int other_agent_num;
		if(mg.a1 == this)	//seller
			other_agent_num = mg.a2->agentNumber;
		else
			other_agent_num = mg.a1->agentNumber;

		//ask questions of all of the other agents
		for(int i = 0; i < mge->agents->size(); i++)
		{
			//don't ask self or the agent playing against
			if(i == agentNumber || i == other_agent_num)
				continue;

			//don't ask an agent that don't know much about
			if(observationHistory[i].observations.size() < minNumObservationsToJudge)
				continue;

			//make game with other agent in it, ask its response
			MarketGame q = mg;
			MarketAgentAction maa;
			if(mg.a1 == this)
			{
				q.a1 = (*mge->agents)[i];
				maa = ((ADFDAgent *)q.a1)->ChooseActionFromPreviousCommunication(q, true);
				//remember results, and record MKT_ACTION_N for the other agent (that was questioned about)
				q.a1Action = maa;
				q.a2Action = MKT_ACTION_N;
			}
			else
			{
				q.a2 = (*mge->agents)[i];
				maa = ((ADFDAgent *)q.a2)->ChooseActionFromPreviousCommunication(q, true);
				//remember results, and record MKT_ACTION_N for the other agent (that was questioned about)
				q.a1Action = MKT_ACTION_N;
				q.a2Action = maa;
			}

			//if the agent doesn't say no, then record the game
			if(maa != MKT_ACTION_N)
				questions.push_back(q);
		}

		//compute action from everything other than new communications
		return ChooseActionFromPreviousCommunication(mg, false);
	}

	//if answering_question is true, then can return FAV_ACTION_N when don't know
	MarketAgentAction ChooseActionFromPreviousCommunication(MarketGame &mg, bool answering_question)
	{
		//find other agent's number
		int other_agent_num;
		if(mg.a1 == this)	//seller
			other_agent_num = mg.a2->agentNumber;
		else
			other_agent_num = mg.a1->agentNumber;

		//only answer if have more than 2 observations
		if(answering_question && observationHistory[other_agent_num].observations.size() < minNumObservationsToJudge)
			return MKT_ACTION_N;

		//create temporary history of this agent based communication
		std::vector<ObservationHistory> observation_history;
		DeepCopyObservationHistory(observationHistory, observation_history);

		//work through agent communications in the order they were received
		for(std::vector<MarketGame>::iterator g = questions.begin(); g != questions.end(); g++)
		{
			//should only judge others when have enough information for both agents in the game in order to judge
			if(observation_history[g->a1->agentNumber].observations.size() < minNumObservationsToJudge
					|| observation_history[g->a2->agentNumber].observations.size() < minNumObservationsToJudge)
				continue;

			//compute discount factor as other agent is seeing it
			double discount_factor_sell_agent = observation_history[g->a1->agentNumber].GetExpectedValue(1 - g->a1->replacementRate, mg.time);
			double discount_factor_buy_agent = observation_history[g->a2->agentNumber].GetExpectedValue(1 - g->a2->replacementRate, mg.time);

			//compute discount factor observation from game
			double seller_discount_factor_observation, buyer_discount_factor_observation;
			ComputeGammaStar(*g, seller_discount_factor_observation, buyer_discount_factor_observation);

			//see what the new action would be by both agents
			MarketAgentAction new_action = MKT_ACTION_R;
			if(discount_factor_sell_agent >= seller_discount_factor_observation
						&& discount_factor_buy_agent >= buyer_discount_factor_observation)
				new_action = MKT_ACTION_P;

			//a1Action or a2Action contains the answer given in the hypothetical game between a1 and a2 (as long as not R)
			//see if the active agent would still play the same answer against a2, given current knowledge
			MarketAgentAction old_action = MKT_ACTION_N;
			if(g->a1Action != MKT_ACTION_N) //seller chose
				old_action = g->a1Action;
			if(g->a2Action != MKT_ACTION_N)
				old_action = g->a2Action;

			//if so, make normal observations for both a1 and a2
			if(new_action == old_action)
			{
				//if affirmitive trust, then record a positive observation for both
				// if some agent a3 says it would not trust agent a2, then we can't infer whether 
				//  this meant that a3 has a low discount factor, a2 has a low discount factor, or both,
				//  so an observation can't be made in that case (unless we find out the agent lied)
				if(new_action == MKT_ACTION_P)
				{
					observation_history[g->a1->agentNumber].AddObservation(seller_discount_factor_observation, 1.0, g->time);
					observation_history[g->a1->agentNumber].RemoveInconsistantObservations();

					observation_history[g->a2->agentNumber].AddObservation(buyer_discount_factor_observation, 1.0, g->time);
					observation_history[g->a2->agentNumber].RemoveInconsistantObservations();
				}
			}
			else //the agent lied about whether it would play the action it did - only record negative observations about the lying agent
			{
				int agent_number;
				double discount_factor_observation;
				if(g->a1Action != MKT_ACTION_N) //seller gave answer
				{
					agent_number = g->a1->agentNumber;
					discount_factor_observation = seller_discount_factor_observation;
				}
				if(g->a2Action != MKT_ACTION_N) //buyer gave answer
				{
					agent_number = g->a1->agentNumber;
					discount_factor_observation = buyer_discount_factor_observation;
				}

				//if recommended playing R, then assume discount factor is below that required for the game
				if(old_action == MKT_ACTION_R)
				{
					observation_history[g->a1->agentNumber].AddObservation(0.0, discount_factor_observation, g->time);
					observation_history[g->a1->agentNumber].RemoveInconsistantObservations();
				}
				else //if recommended playing P, then make a "really low" observation
				{
					observation_history[g->a1->agentNumber].AddObservation(0.0, .05, g->time);
					observation_history[g->a1->agentNumber].RemoveInconsistantObservations();
				}
			}
		}

		//this represents the observation that will be made for this game
		double seller_discount_factor_observation, buyer_discount_factor_observation;
		ComputeGammaStar(mg, seller_discount_factor_observation, buyer_discount_factor_observation);

		MarketAgent *other_agent;
		double self_discount_factor_observation, other_discount_factor_observation;

		double immediate_value_of_playing_P, immediate_value_of_playing_N;

		if(mg.a1 == this)	//seller
		{
			other_agent_num = mg.a2->agentNumber;
			other_agent = mg.a2;
			double buyer_discount_factor_estimate = observationHistory[other_agent_num].GetExpectedValue(1 - other_agent->replacementRate, mg.time);

			//protect self from other playing N
			if(buyer_discount_factor_estimate < buyer_discount_factor_observation)
				return MKT_ACTION_R;

			self_discount_factor_observation = seller_discount_factor_observation;
			other_discount_factor_observation = buyer_discount_factor_observation;

			immediate_value_of_playing_P = mg.m - mg.c;
			immediate_value_of_playing_N = mg.m;
		}
		else //buyer
		{
			other_agent_num = mg.a1->agentNumber;
			other_agent = mg.a1;
			double seller_discount_factor_estimate = observationHistory[other_agent_num].GetExpectedValue(1 - other_agent->replacementRate, mg.time);

			//protect self from other playing N
			if(seller_discount_factor_estimate < seller_discount_factor_observation)
				return MKT_ACTION_R;

			self_discount_factor_observation = buyer_discount_factor_observation;
			other_discount_factor_observation = seller_discount_factor_observation;

			immediate_value_of_playing_P = mg.w - mg.m;
			immediate_value_of_playing_N = mg.w;
		}

		//compute utilities of playing different actions based on how it will affect other agents' perceptions
		std::vector<int> agents;	agents.push_back(other_agent_num);

		//compute utility given playing P
		std::vector<ObservationHistory> observation_self_history;
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(self_discount_factor_observation, 1, mg.time);
		observation_self_history[other_agent_num].RemoveInconsistantObservations();

		double V_P = ComputeContinuationValue(discountFactor, agents, observation_history, observation_self_history, mg.time+1);
		V_P += immediate_value_of_playing_P;

		//compute utility given playing R
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(0, self_discount_factor_observation, mg.time);
		observation_self_history[other_agent_num].RemoveInconsistantObservations();

		double V_N = ComputeContinuationValue(discountFactor, agents, observation_history, observation_self_history, mg.time+1);
		V_N += immediate_value_of_playing_N;

		//no change to its own self history by playing R
		double V_R = ComputeContinuationValue(discountFactor, agents, observation_history, observationSelfHistory, mg.time+1);

		if(V_P > V_N && V_P > V_R)
			return MKT_ACTION_P;
		if(!answering_question && V_N > V_R)	//don't want to tell other that this agent would play N (which is reserved for another meaning anyway)
			return MKT_ACTION_N;
		return MKT_ACTION_R;
	}


	MarketAgentAction ChooseActionFromExpectedUtilityChanges(MarketGame &mg)
	{
		//this represents the observation that will be made for this game
		double seller_discount_factor_observation, buyer_discount_factor_observation;
		ComputeGammaStar(mg, seller_discount_factor_observation, buyer_discount_factor_observation);

		int other_agent_num;
		MarketAgent *other_agent;
		double self_discount_factor_observation, other_discount_factor_observation;

		double immediate_value_of_playing_P, immediate_value_of_playing_N;

		if(mg.a1 == this)	//seller
		{
			other_agent_num = mg.a2->agentNumber;
			other_agent = mg.a2;
			double buyer_discount_factor_estimate = observationHistory[other_agent_num].GetExpectedValue(1 - other_agent->replacementRate, mg.time);

			//protect self from other playing N
			if(buyer_discount_factor_estimate < buyer_discount_factor_observation)
				return MKT_ACTION_R;

			self_discount_factor_observation = seller_discount_factor_observation;
			other_discount_factor_observation = buyer_discount_factor_observation;

			immediate_value_of_playing_P = mg.m - mg.c;
			immediate_value_of_playing_N = mg.m;
		}
		else //buyer
		{
			other_agent_num = mg.a1->agentNumber;
			other_agent = mg.a1;
			double seller_discount_factor_estimate = observationHistory[other_agent_num].GetExpectedValue(1 - other_agent->replacementRate, mg.time);

			//protect self from other playing N
			if(seller_discount_factor_estimate < seller_discount_factor_observation)
				return MKT_ACTION_R;

			self_discount_factor_observation = buyer_discount_factor_observation;
			other_discount_factor_observation = seller_discount_factor_observation;

			immediate_value_of_playing_P = mg.w - mg.m;
			immediate_value_of_playing_N = mg.w;
		}

		//compute utilities of playing different actions based on how it will affect other agents' perceptions
		std::vector<int> agents;	agents.push_back(other_agent_num);

		//compute utility given playing P
		std::vector<ObservationHistory> observation_self_history;
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(self_discount_factor_observation, 1, mg.time);
		observation_self_history[other_agent_num].RemoveInconsistantObservations();

		double V_P = ComputeContinuationValue(discountFactor, agents, observationHistory, observation_self_history, mg.time+1);
		V_P += immediate_value_of_playing_P;

		//compute utility given playing R
		DeepCopyObservationHistory(observationSelfHistory, observation_self_history);

		observation_self_history[other_agent_num].AddObservation(0, self_discount_factor_observation, mg.time);
		observation_self_history[other_agent_num].RemoveInconsistantObservations();

		double V_N = ComputeContinuationValue(discountFactor, agents, observationHistory, observation_self_history, mg.time+1);
		V_N += immediate_value_of_playing_N;

		//no change to its own self history by playing R
		double V_R = ComputeContinuationValue(discountFactor, agents, observationHistory, observationSelfHistory, mg.time+1);

		if(V_P > V_N && V_P > V_R)
			return MKT_ACTION_P;
		if(V_N > V_R)
			return MKT_ACTION_N;
		return MKT_ACTION_R;
	}

	MarketAgentAction NaiivelyChooseAction(MarketGame &mg)
	{
		if(mg.a1 == this)	//sell agent
		{
			double expected_opponent_discount_factor = observationHistory[mg.a2->agentNumber].GetExpectedValue(1 - mg.a2->replacementRate, mg.time);
			return GetSafeAction(mg, discountFactor, expected_opponent_discount_factor);
		}
		else //buy agent
		{
			double expected_opponent_discount_factor = observationHistory[mg.a1->agentNumber].GetExpectedValue(1 - mg.a1->replacementRate, mg.time);
			return GetSafeAction(mg, expected_opponent_discount_factor, discountFactor);
		}
	}

	virtual void MakeObservation(MarketGame &mg)
	{
		//want to determine the discount factor boundary that this game represents
		double seller_discount_factor_estimate, buyer_discount_factor_estimate;
		ComputeGammaStar(mg, seller_discount_factor_estimate, buyer_discount_factor_estimate);

		//add observation
		if(mg.a1 == this)	//sell agent
		{
			if(mg.a1Action == MKT_ACTION_P)
				observationSelfHistory[mg.a2->agentNumber].AddObservation(seller_discount_factor_estimate, 1.0, mg.time);
			else if(mg.a1Action == MKT_ACTION_N)
				observationSelfHistory[mg.a2->agentNumber].AddObservation(0.0, seller_discount_factor_estimate, mg.time);
			if(mg.a1Action != MKT_ACTION_R)
				observationSelfHistory[mg.a2->agentNumber].RemoveInconsistantObservations();

			if(mg.a2Action == MKT_ACTION_P)
				observationHistory[mg.a2->agentNumber].AddObservation(buyer_discount_factor_estimate, 1.0, mg.time);
			else if(mg.a2Action == MKT_ACTION_N)
				observationHistory[mg.a2->agentNumber].AddObservation(0.0, buyer_discount_factor_estimate, mg.time);
			if(mg.a2Action != MKT_ACTION_R)
				observationHistory[mg.a2->agentNumber].RemoveInconsistantObservations();
		}
		else if(mg.a2 == this)	//buy agent
		{	
			if(mg.a1Action == MKT_ACTION_P)
				observationSelfHistory[mg.a2->agentNumber].AddObservation(buyer_discount_factor_estimate, 1.0, mg.time);
			else if(mg.a1Action == MKT_ACTION_N)
				observationSelfHistory[mg.a2->agentNumber].AddObservation(0.0, buyer_discount_factor_estimate, mg.time);
			if(mg.a1Action != MKT_ACTION_R)
				observationSelfHistory[mg.a2->agentNumber].RemoveInconsistantObservations();

			if(mg.a2Action == MKT_ACTION_P)
				observationHistory[mg.a2->agentNumber].AddObservation(seller_discount_factor_estimate, 1.0, mg.time);
			else if(mg.a2Action == MKT_ACTION_N)
				observationHistory[mg.a2->agentNumber].AddObservation(0.0, seller_discount_factor_estimate, mg.time);
			if(mg.a2Action != MKT_ACTION_R)
				observationHistory[mg.a2->agentNumber].RemoveInconsistantObservations();
		}
	}


	int minNumObservationsToJudge;
	
	//this agent's observation of other agents
	std::vector<ObservationHistory> observationHistory;
	//other agents' observations of this agent
	std::vector<ObservationHistory> observationSelfHistory;

	//stores a list of games for each pair of agents
	// say a1 is deciding whether to offer a2 a favor, and asks a3 if a3 would offer a favor in the current game to a2
	//  questions[a3][a2] would contain a vector of the games of which a1 asked a3 a "question" about a2
	// the game's a1Action or a2Action field contains the "answer" from the agent the question was asked
	// and the other field contains MKT_ACTION_R
	std::vector<MarketGame> questions;
};
