enum MarketAgentAction
{
	MKT_ACTION_P,	//positive (pay)
	MKT_ACTION_N,	//negative (for communication: no comment)
	MKT_ACTION_R	//refrain
};

class MarketAgent;

//used to describe a single market game to be passed into an agent
class MarketGame
{
public:
	MarketAgent *a1, *a2;
	MarketAgentAction a1Action, a2Action;
	double c, m, w;
	double time;
};

class MarketAgent : public Agent
{
public:
	MarketAgent(int agent_number, int num_agents, RandomVariable *discount_factor_distribution, double replacement_rate)
		: Agent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{

	}
	
	//if agent is a1, then is selling, if a2, then buying
	virtual MarketAgentAction ChooseAction(MarketGame &mg)
	{
		double r = genrand_real1();
		if(r <= 1.0/3)
			return MKT_ACTION_R;
		else if(r <= 2.0/3)
			return MKT_ACTION_N;
		else
			return MKT_ACTION_P;
	}

	virtual void MakeObservation(MarketGame &mg)
	{	}

protected:

};

class MarketGameEnvironment : public MultiAgentSimulation<MarketAgent, MarketGame>
{
public:
	MarketGameEnvironment(RandomVariable *c, RandomVariable *m, RandomVariable *w, std::vector<MarketAgent *> *agent_list)
	{
		C = c; M = m; W = w;
		agents = agent_list;
	}
	
	void AssignPayoffs(MarketGame &mg)
	{
		//no information gained if either played R
		if(mg.a1Action == MKT_ACTION_R || mg.a2Action == MKT_ACTION_R)
			return;

		if(mg.a1Action == MKT_ACTION_P)
		{
			if(mg.a2Action == MKT_ACTION_P)
			{
				mg.a1->totalUtility += mg.m - mg.c;
				mg.a2->totalUtility += mg.w - mg.m;
			}
			else //mg.a2Action == MKT_ACTION_N
			{
				mg.a1->totalUtility += -mg.c;
				mg.a2->totalUtility += mg.w;
			}			
		}
		else //mg.a1Action == MKT_ACTION_N
		{
			if(mg.a2Action == MKT_ACTION_P)
			{
				mg.a1->totalUtility += mg.m;
				mg.a2->totalUtility += -mg.m;
			}
			//if mg.a2Action == MKT_ACTION_N, then both get a payoff of 0
		}
	}
	
	void PrintGameInfo(MarketGame &mg)
	{
		char a1, a2;
		switch(mg.a1Action)
		{
			case MKT_ACTION_R: a1 = 'R';	break;
			case MKT_ACTION_P: a1 = 'P';	break;
			case MKT_ACTION_N: a1 = 'N';	break;
		}
		switch(mg.a2Action)
		{
			case MKT_ACTION_R: a2 = 'R';	break;
			case MKT_ACTION_P: a2 = 'P';	break;
			case MKT_ACTION_N: a2 = 'N';	break;
		}
		printf(" a1:%d:%c a2:%d:%c c:%f m:%f w:%f\n", mg.a1->agentNumber, a1, mg.a2->agentNumber, a2,
			mg.c, mg.m, mg.w);
	}
	
	bool ExecuteRound(double cur_time)
	{
		//check for discount factor replacements
//		for(int i = 0; i < agents->size(); i++)
//			(*agents)[i]->UpdateDiscountFactor(cur_time);

		//for every combination of agents as buyer and seller, see if should play a game
		for(int i = 0; i < agents->size(); i++)
			for(int j = 0; j < agents->size(); j++)
			{
				//an agent can't play itself
				if(i == j)
					continue;

				if(genrand_real1() >= (*agents)[i]->encounterRate[j])
					continue;

				//build a game to play
				MarketGame mg;
				mg.a1 = (*agents)[i];
				mg.a2 = (*agents)[j];

				mg.c = C->GetNextRand();
				mg.m = M->GetNextRand();
				mg.w = W->GetNextRand();

				mg.time = cur_time;

				//get agents' moves
				mg.a1Action = mg.a1->ChooseAction(mg);
				mg.a2Action = mg.a2->ChooseAction(mg);

				//store results and payoffs
				mg.a1->MakeObservation(mg);
				mg.a2->MakeObservation(mg);

				AssignPayoffs(mg);

				PrintGameInfo(mg);
			}

		return false;	//the infinite game is never done
	}

	RandomVariable *C, *M, *W;
};
