
extern RobotFinDeliveryGameEnvironment *rfdge;

class RobotFinDeliveryProbabilisticReciprocityAgent : public RobotFinDeliveryAgent
{
public:
	RobotFinDeliveryProbabilisticReciprocityAgent(int agent_number, int num_agents, RandomVariable *distance_distro, RandomVariable *discount_factor_distribution, double replacement_rate)
		: RobotFinDeliveryAgent(agent_number, num_agents, distance_distro, discount_factor_distribution, replacement_rate)
	{
		agentBallance.resize(num_agents, 0);

	}

	virtual bool ChooseAction(RobotFinDeliveryGame &rfdg)
	{
		//return true;

		double tau = 0.75;
		double beta = 0.5;
		double average_task_cost = taskDistanceDistribution->GetExpectedValue();
		double current_cost = rfdg.askeeTaskCost;
		double ballance = agentBallance[rfdg.askee->agentNumber];

		double pr = 1.0 / (1.0 + exp(
									(current_cost - beta * average_task_cost - ballance) / tau ));

		double r = genrand_real1();
		if(r < pr)
			return true;
		return false;
	}

	virtual void MakeObservation(RobotFinDeliveryGame &rfdg)
	{	
		if(rfdg.asker == this)
		{
			if(rfdg.askeePerformedFavor)
			{
				int other_agent_num = rfdg.askee->agentNumber;
				agentBallance[other_agent_num] += rfdg.askerTaskCost;
			}

		}
		else //was askee
		{
			if(rfdg.askeePerformedFavor)
			{
				int other_agent_num = rfdg.asker->agentNumber;
				agentBallance[other_agent_num] -= rfdg.askeeTaskCost;
			}

		}
	}

	//the index is the agent number that has the ballance
	std::vector<int> agentBallance;
};

class RobotFinDeliveryDiscountFactorMeasuringAgent : public RobotFinDeliveryAgent
{
public:
	RobotFinDeliveryDiscountFactorMeasuringAgent(int agent_number, int num_agents, RandomVariable *distance_distro, RandomVariable *discount_factor_distribution, double replacement_rate, int min_num_observations_to_judge)
		: RobotFinDeliveryAgent(agent_number, num_agents, distance_distro, discount_factor_distribution, replacement_rate)
	{
		for(int i = 0; i < num_agents; i++)
		{
			observationHistory.push_back(ObservationHistory());
			observationSelfHistory.push_back(ObservationHistory());
		}

		minNumObservationsToJudge = min_num_observations_to_judge;
	}

	virtual bool ChooseAction(RobotFinDeliveryGame &rfdg)
	{
		return true;
	}

	virtual void MakeObservation(RobotFinDeliveryGame &rfdg)
	{	
		if(rfdg.asker == this)
		{
			if(rfdg.askeePerformedFavor)
			{
				
			}

		}
		else //was askee
		{
			if(rfdg.askeePerformedFavor)
			{
				
			}

		}
	}
	
	virtual void PrintInfo()
	{
		printf("agent %d:\t$: %f\tdf: %f\tr: %f\n", agentNumber, totalUtility, discountFactor, replacementRate);
	}	
	
	int minNumObservationsToJudge;

	//this agent's direct observations of other agents
	std::vector<ObservationHistory> observationHistory;
	//other agents' direct observations of this agent
	std::vector<ObservationHistory> observationSelfHistory;

	//stores a list of games for each pair of agents
	// say a1 is deciding whether to offer a2 a favor, and asks a3 if a3 would offer a favor in the current game to a2
	//  questions[a3][a2] would contain a vector of the games of which a1 asked a3 a "question" about a2
	// the game's a1Action field contains the "answer"
	std::vector<FavorGame> questions;
};
