
class RobotFinDeliveryAgent;

class RobotFinDelivery
{
public:
	RobotFinDeliveryAgent *taskOwner;
	int radialFin;
	int distance;	//distance asker's package needs to go
};

//used to describe a single favor game to be passed into an agent
class RobotFinDeliveryGame
{
public:
	RobotFinDeliveryAgent *asker, *askee;
	int radialFin;
	int distance;	//distance asker currently needs to go
	int askerTaskCost;
	int askeeTaskCost;
	bool askeePerformedFavor;	//true if askee performed the favor asked
	double time;	//current time
};

class RobotFinDeliveryAgent : public Agent
{
public:
	RobotFinDeliveryAgent(int agent_number, int num_agents, RandomVariable *distance_distro, RandomVariable *discount_factor_distribution, double replacement_rate)
		: Agent(agent_number, num_agents, discount_factor_distribution, replacement_rate)
	{
		travelDistance = 0;
		totalTravelDistance = 0;
		currentTaskCost = 0;
		busyUntilTime = 0.0;
		numTasksCompleted = 0;
		taskDistanceDistribution = distance_distro;
	}
	
	virtual bool ChooseAction(RobotFinDeliveryGame &rfdg)
	{
		return true;
		double r = genrand_real1();
		if(r < 0.5)
			return true;
		else
			return false;
	}

	int GetMarginalCostOfTask(RobotFinDelivery &rfd)
	{
		if(rfd.distance <= travelDistance)
			return rfd.distance;	//extra distance to take along path
		else
			return rfd.distance + (rfd.distance - travelDistance);
	}

	void AssignTask(RobotFinDelivery &rfd)
	{
		currentRadialFin = rfd.radialFin;
		currentTaskCost += GetMarginalCostOfTask(rfd);
		travelDistance = std::max(travelDistance, rfd.distance);
	}

	virtual void MakeObservation(RobotFinDeliveryGame &rfdg)
	{	}

	void CommitCurrentTasks(double cur_time)
	{
		totalTravelDistance += 2 * travelDistance;
		totalUtility += currentTaskCost;
		busyUntilTime += 2 * travelDistance;
		busyUntilTime = std::max(busyUntilTime, cur_time);
		travelDistance = 0;
		currentTaskCost = 0;
	}

	virtual void PrintInfo()
	{
		printf("agent %d:\t$: %.03f\tdf: %.03f  r: %.03f\td: %d\tt: %d\n",
				agentNumber, totalUtility, discountFactor, replacementRate, totalTravelDistance, numTasksCompleted);
	}

	int travelDistance;
	int totalTravelDistance;
	int currentTaskCost;
	double busyUntilTime;
	int currentRadialFin;

	int numTasksCompleted;

	RandomVariable *taskDistanceDistribution;

protected:

};

//current implementation based on paper "Believing Others: Pros and Cons"

class RobotFinDeliveryGameEnvironment : public MultiAgentSimulation<RobotFinDeliveryAgent, RobotFinDeliveryGame>
{
public:
	RobotFinDeliveryGameEnvironment(RandomVariable *fin_number_distro, int num_robot_tasks, std::vector<RobotFinDeliveryAgent *> *agent_list)
	{
		radialfinDistribution = fin_number_distro;
		numRobotTasks = num_robot_tasks;
		agents = agent_list;
	}

	void PrintGameInfo(RobotFinDeliveryGame &rfdg)
	{
		printf(" asker:%d askee:%d:%c %d:%d (%d)\n", rfdg.asker->agentNumber, rfdg.askee->agentNumber,
			(rfdg.askeePerformedFavor ? 'y' : 'n'), rfdg.askerTaskCost, rfdg.askeeTaskCost, rfdg.distance);
	}

	bool ExecuteRound(double cur_time)
	{
		//get random order
		//make proper amount of numbers in order
		std::vector<int> order;
		order.reserve(agents->size());
		for(int i = 0; i < agents->size(); i++)
			order.push_back(i);
		
		//shuffle numbers
		RandomVariable *random_sample = new UniformDistribution(0, 1);
		int i = agents->size();
		while(--i > 0) 
		{
			int k = (int)(random_sample->GetNextRand() * (i + 1));  // [0,i]
			int tmp = order[i];
			order[i] = order[k];
			order[k] = tmp;
		}
		delete random_sample;
		
		bool robot_with_no_task = true;
		bool any_robot_not_done = false;

		while(robot_with_no_task)
		{
			robot_with_no_task = false;

			//assign every agent at the base a new task
			for(int i = 0; i < order.size(); i++)
			{
				//get next robot to assign task to
				RobotFinDeliveryAgent *rfda = (*agents)[order[i]];

				//see if robot is done
				if(rfda->numTasksCompleted >= numRobotTasks)
					continue;

				//at least one robot isn't done
				any_robot_not_done = true;

				//if robot is still working on a task, can't assign it a new one
				if(rfda->busyUntilTime > cur_time)
					continue;

				//if robot has already been assigned a task, don't give it a new one
				if(rfda->travelDistance > 0)
					continue;

				//create a new task
				RobotFinDelivery rfd;
				rfd.distance = (int)rfda->taskDistanceDistribution->GetNextRand();
				rfd.radialFin = (int)(radialfinDistribution->GetNextRand());
				rfd.taskOwner = rfda;

				//count the task
				rfda->numTasksCompleted++;
				int task_cost = rfda->GetMarginalCostOfTask(rfd);

				bool performed_by_another_agent = false;

				//check every other agent to see if it will do the task
				for(int j = 0; j < i; j++)
				{
					//an agent can't ask itself
					if(i == j)
						continue;

					RobotFinDeliveryAgent *askee = (*agents)[order[j]];
					//if the askee doesn't have a destination yet, then can't ask it for a favor
					if(askee->travelDistance == 0)
						continue;

					//need to make sure that this agent is going to the correct radial fin
					if(askee->currentRadialFin != rfd.radialFin)
						continue;

					//if the other agent can't do the task for cheaper, then don't ask it
					int askee_task_cost = askee->GetMarginalCostOfTask(rfd);
					if(task_cost <= askee_task_cost)
						continue;

					//build a game to play
					RobotFinDeliveryGame rfdg;
					rfdg.asker = rfda;
					rfdg.askee = (*agents)[j];
					rfdg.askerTaskCost = task_cost;
					rfdg.askeeTaskCost = askee_task_cost;
					rfdg.distance = rfd.distance;
					rfdg.radialFin = rfd.radialFin;
					rfdg.time = cur_time;

					rfdg.askeePerformedFavor = askee->ChooseAction(rfdg);

					PrintGameInfo(rfdg);

					rfda->MakeObservation(rfdg);
					askee->MakeObservation(rfdg);

					if(rfdg.askeePerformedFavor)
					{
						askee->AssignTask(rfd);
						robot_with_no_task = true;
						performed_by_another_agent = true;
						break;
					}
				} //for all askees

				if(!performed_by_another_agent)
					rfda->AssignTask(rfd);
			} //for all agents

		} //while !robot_with_no_task

		//committ tasks
		//assign every agent at the base a new task
		for(int i = 0; i < order.size(); i++)
		{
			//if robot is still working on a task, skip it
			RobotFinDeliveryAgent *rfda = (*agents)[order[i]];
			rfda->CommitCurrentTasks(cur_time);
		}

		return any_robot_not_done;
	}

	RandomVariable *radialfinDistribution;
	int numRobotTasks;
};
