//////////////////////////

RandomVariable *c, *w;
double r[2];

class MaxFavorEvaluator : public Evaluator
{
public:
	double gamma;
	int agentNum;
	double Evaluate(std::vector<double> v)
	{
		int other_agent = 1-agentNum;
		double c_a = v[agentNum];
		double c_other = v[other_agent];

		return gamma * r[agentNum] * c->GetProbabilityInRange(-INFINITY, c_other) * w->GetExpectedValue()
			- gamma * r[other_agent] * c->GetProbabilityInRange(-INFINITY, c_a) * c->GetExpectedValueInRange(-INFINITY, c_a)
			- (1-gamma) * c_a;
	}
};

void WriteOutGammaRanges()
{
	c = new UniformDistribution(0, 200);
	w = new UniformDistribution(550, 650);

	MaxFavorEvaluator t1;
	MaxFavorEvaluator t2;
	std::vector<Evaluator *> tests;
	tests.push_back(&t1);
	tests.push_back(&t2);

	r[0] = .14;
	t1.gamma = .8;
	t1.agentNum = 0;

	r[1] = .2;
	t2.gamma = .84;
	t2.agentNum = 1;

	const double min_c_check = -20;
	const double max_c_check = 15000;

	std::vector<double> min;		min.push_back(min_c_check);		min.push_back(min_c_check);
	std::vector<double> max;		max.push_back(max_c_check);		max.push_back(max_c_check);
	std::vector<double> tolerance;	tolerance.push_back(0.001);		tolerance.push_back(0.001);
	std::vector<double> results;	results.push_back(0);							results.push_back(0);

	const int s = 32;
	double a1[s][s];
	double a2[s][s];
	for(int i = 0; i < s; i++)
		for(int j = 0; j < s; j++)
		{
			t1.gamma = (double)i / s;
			t2.gamma = (double)j / s;
			if(SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
			{
				a1[i][j] = results[0];	a2[i][j] = results[1];
			}
			else
			{
				a1[i][j] = -1;	a2[i][j] = -1;
			}
		}

	std::ofstream f;
	f.open("out.csv");

	//write out headers
	f << "g1\\g2,";
	for(int j = 0; j < s; j++)
		f << (double)j / s << ",";
	f << std::endl;

	for(int i = 0; i < s; i++)
	{
		f << (double)i / s << ",";	//write out row header
		for(int j = 0; j < s; j++)
			f << a1[i][j] << ",";
		f << std::endl;
	}
	f << std::endl;
	for(int i = 0; i < s; i++)
	{
		f << (double)i / s << ",";	//write out row header
		for(int j = 0; j < s; j++)
			f << a2[i][j] << ",";
		f << std::endl;
	}
	f.close();
}

////////////////////////////////

void TestValueObservation()
{
	ValueObservation vo(.3, .7, 0.0);
	for(double x = 0; x <= 1.0; x += 0.1)
		printf("%.03f ", vo.GetPDFValue(x, .7, 0, 0.1));
}

void TestObservationHistory()
{
	ObservationHistory no_observations;

	ObservationHistory oh;
//	oh.AddObservation(0, .5, 0.0);
//	oh.AddObservation(0, .7, 1.0);
//	oh.AddObservation(.3, 1, 5.0);
	
	oh.AddObservation(0, .5, 0.0);
	oh.AddObservation(0, .3, 4.99);
	oh.AddObservation(.3, 1, 5.0);

	oh.probabilityOfError = 0.4;
	printf("oh expected value: %f\n", oh.GetExpectedValue(.7, 5.0));

	double diff_ent = no_observations.GetRelativeEntropyGivenNewObservationHistory(oh, .7, 5.0, 0.1);
	printf("oh diff ent: %f\n", diff_ent);
	
	double best_prob = oh.GetProbabilityOfErrorWithMaximumEntropy(.7, 5.0);
	printf("oh error probability: %f\n", best_prob);

	oh.probabilityOfError = best_prob;
	printf("oh expected value w/ error: %f\n", oh.GetExpectedValue(.7, 5.0));

	ObservationHistory oh1;//(0.1, 2.0);
	oh1.AddObservation(.3, .5, 0.0);
	oh1.AddObservation(.1, .8, 2.0);
	oh1.AddObservation(.2, .5, 1.0);
	oh1.AddObservation(.3, .9, 3.0);
	printf("%f\n", oh1.GetExpectedValue(.7, 6.0));
	oh1.RemoveInconsistantObservations();
	printf("%f\n", oh1.GetExpectedValue(.7, 6.0));

	ObservationHistory oh2;
	oh2.AddObservation(.95, .951, 3.95);
	double h = oh1.GetRelativeEntropyGivenNewObservationHistory(oh2, .7, 4.0);

	printf("entropy: %f\n", h);
	
	oh1.AddObservation(.3, 1.0, 6.0);
	oh1.AddObservation(.0, .7, 2.0);
	oh1.AddObservation(.0, .5, 1.0);
	printf("%f\n", oh1.GetExpectedValue(.7, 6.0));


	ObservationHistory ohc;
//	ohc.AddObservation(0.31263085842132565, 1, 0.0);
//	ohc.AddObservation(0.88199759159088120, 1.0, 1.0);
//	ohc.AddObservation(0, 0.46260212688446045, 2.0);
//	ohc.AddObservation(0, 0.94375899295806887, 3.0);
//	ohc.AddObservation(0, 0.82209757633209224, 4.0);
//	ohc.AddObservation(0, 0.63971988086700426, 8.0);
//	ohc.AddObservation(0, 0.73820170040130606, 9.0);

	ohc.AddObservation(0, 0.5, 0.0);
	ohc.AddObservation(0.5, 1, 1.0);
	ohc.AddObservation(0, 0.5, 2.0);
	ohc.AddObservation(0.5, 1, 3.0);
	ohc.AddObservation(0, 0.5, 4.0);
	ohc.AddObservation(0.5, 1, 5.0);
	ohc.AddObservation(0, 0.5, 6.0);
	ohc.AddObservation(0.5, 1, 7.0);
	ohc.AddObservation(0, 0.5, 8.0);
	ohc.AddObservation(0.5, 1, 9.0);

	ObservationHistory ohc2 = ohc;

	double replacement_rate = .01;

	printf("comparing observation history inconsistancy methods\n");
	printf("ohc initial expected value: %f\n", ohc.GetExpectedValue(replacement_rate, 10.0));

	double next_observation_low = 0;
	double next_observation_high = 0.752; //80370875988006585;
	next_observation_low = .1;
	next_observation_high = 1;

	next_observation_low = 0;
	next_observation_high = .5;

	ohc.HandleInconsistantObservations(replacement_rate, 10, true);
	printf("ohc expected value removed inconsistent observations: %f\n", ohc.GetExpectedValue(replacement_rate, 10));
	ohc.AddObservation(next_observation_low, next_observation_high, 10);
	printf("ohc expected value 2: %f\n", ohc.GetExpectedValue(replacement_rate, 11));
	ohc.HandleInconsistantObservations(replacement_rate, 11, true);
	printf("ohc expected value 2 removed inconsistent observations: %f\n", ohc.GetExpectedValue(replacement_rate, 11));

	ohc2.HandleInconsistantObservations(replacement_rate, 10, false);
	printf("ohc2 expected value removed prob error: %f   (prob err: %f)\n", ohc2.GetExpectedValue(replacement_rate, 10), ohc2.probabilityOfError);
	ohc2.AddObservation(next_observation_low, next_observation_high, 10);
	printf("ohc2 expected value 2: %f\n", ohc2.GetExpectedValue(replacement_rate, 11));
	ohc2.HandleInconsistantObservations(replacement_rate, 11, false);
	printf("ohc2 expected value 2 removed inconsistent observations: %f   (prob err: %f)\n", ohc2.GetExpectedValue(replacement_rate, 11), ohc2.probabilityOfError);
}

void TestFavorGammaEvaluator()
{
	c = new UniformDistribution(1, 5);
	w = new UniformDistribution(10, 30);

	MaxFavorEvaluator t1;
	MaxFavorEvaluator t2;
	std::vector<Evaluator *> tests;
	tests.push_back(&t1);
	tests.push_back(&t2);

	r[0] = .14;
	t1.gamma = .8;
	t1.agentNum = 0;

	r[1] = .2;
	t2.gamma = .84;
	t2.agentNum = 1;

	std::vector<double> min;
	min.push_back(1);
	min.push_back(1);

	std::vector<double> max;
	max.push_back(5);
	max.push_back(5);

	std::vector<double> tolerance;	tolerance.push_back(0.001);							tolerance.push_back(.001);

	std::vector<double> results;
	results.push_back(0);
	results.push_back(0);


	if(SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
		for(int i = 0; i < results.size(); i++)
			printf("%f ", results[i]);
	else
		printf("no solution");
}


void TestADFDComputeCMStarAndComputeGammaStar()
{
	MarketGame mg;
	mg.a1 = (*mge->agents)[0];
	mg.a2 = (*mge->agents)[1];
	mg.a1->encounterRate[1] = .7;
	mg.a2->encounterRate[0] = .7;
	double c_a1, m_a2;

	if(ADFDAgent::ComputeCMStar(mg, .5, .56, c_a1, m_a2))
	{
		printf("%f, %f", c_a1, m_a2);
	}

	mg.m = m_a2;
	mg.c = c_a1;
	double g1, g2;
	ADFDAgent::ComputeGammaStar(mg, g1, g2);
	return;
}

void WriteOutCMRanges()
{
	MarketGame mg;
	mg.a1 = (*mge->agents)[0];
	mg.a2 = (*mge->agents)[1];
	mg.a1->encounterRate[1] = .7;
	mg.a2->encounterRate[0] = .7;
	double c_a1, m_a2;

	const int s = 32;
	double a1[s][s];
	double a2[s][s];
	for(int i = 0; i < s; i++)
		for(int j = 0; j < s; j++)
		{
			double g1 = (double)i / s;
			double g2 = (double)j / s;
			if(ADFDAgent::ComputeCMStar(mg, g1, g2, c_a1, m_a2))
			{
				a1[i][j] = c_a1;	a2[i][j] = m_a2;
			}
			else
			{
				a1[i][j] = -1;	a2[i][j] = -1;
			}
		}

	std::ofstream f;
	f.open("out.csv");
	for(int i = 0; i < s; i++)
	{
		for(int j = 0; j < s; j++)
			f << a1[i][j] << ",";
		f << std::endl;
	}
	f << std::endl;
	for(int i = 0; i < s; i++)
	{
		for(int j = 0; j < s; j++)
			f << a2[i][j] << ",";
		f << std::endl;
	}
	f.close();
}

void TestBasicMarketGame()
{
	double replacement_rate = 0.02;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);
	
	int num_agents = 2;
	std::vector<MarketAgent *> *agent_list = new std::vector<MarketAgent *>();
	agent_list->reserve(num_agents);
	for(int i = 0; i < num_agents; i++)
		agent_list->push_back(new ADFDAgent(i, num_agents, discount_factor_distro, replacement_rate, 4));

	RandomVariable *c, *m, *w;
	c = new UniformDistribution(1, 2);
	m = new UniformDistribution(3, 4);
	w = new UniformDistribution(5, 6);
	mge = new MarketGameEnvironment(c, m, w, agent_list);

	mge->RunSimulation(500);
}

void TestFavorGammaC()
{
	double replacement_rate = 0.02;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);
	RandomVariable *agent_interaction_rate = new UniformDistribution(0.0, 1, false);
	int communication_min_observations = 200;

	std::vector<FavorAgent *> *favor_agent_list = new std::vector<FavorAgent *>();

	FavorAgent *fa1 = new CommunicatingReciprocatingFavorAgent(0, 2, discount_factor_distro, replacement_rate, communication_min_observations, true, FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL);
	favor_agent_list->push_back(fa1);

	FavorAgent *fa2 = new CommunicatingReciprocatingFavorAgent(1, 2, discount_factor_distro, replacement_rate, communication_min_observations, true, FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL);
	favor_agent_list->push_back(fa2);

	fa1->encounterRate[1] = 0.75;
	fa2->encounterRate[0] = 0.75;

	RandomVariable *c, *w;
//	c = new UniformDistribution(0, 100);
//	c = new ExponentialDistribution(100);
//	w = new UniformDistribution(50, 150);

//	c = new UniformDistribution(0, 200);
	c = new ExponentialDistribution(100);
	w = new UniformDistribution(550, 650);
	fge = new FavorGameEnvironment(c, w, favor_agent_list);


	FavorGame fg;
	fg.c = 100;
	fg.w = w->GetExpectedValue();
	fg.a1 = fa1;
	fg.a2 = fa2;

	const int s = 32;
	double a1[s][s];
	double a2[s][s];
	for(int i = 0; i < s; i++)
		for(int j = 0; j < s; j++)
		{
			fa1->discountFactor = (double)i / s;
			fa2->discountFactor = (double)j / s;

			double c_a1, c_a2;
			//if(ComputeCStar<MaxCGammaValueFavorEvaluatorRational>(fg, fa1->discountFactor, fa2->discountFactor, c_a1, c_a2))
			//if(ComputeCStar<MaxCGammaValueFavorEvaluatorDerivative>(fg, fa1->discountFactor, fa2->discountFactor, c_a1, c_a2))
			if(ComputeCStar<MaxCGammaValueFavorEvaluatorSanction>(fg, fa1->discountFactor, fa2->discountFactor, c_a1, c_a2))
			{
				a1[i][j] = c_a1;	a2[i][j] = c_a2;
			}
			else
			{
				a1[i][j] = -1;	a2[i][j] = -1;
			}
		}

	std::ofstream f;
	f.open("out.csv");

	//write out headers
	f << "g1\\g2,";
	for(int j = 0; j < s; j++)
		f << (double)j / s << ",";
	f << std::endl;

	for(int i = 0; i < s; i++)
	{
		f << (double)i / s << ",";	//write out row header
		for(int j = 0; j < s; j++)
			f << a1[i][j] << ",";
		f << std::endl;
	}
	f << std::endl;
	for(int i = 0; i < s; i++)
	{
		f << (double)i / s << ",";	//write out row header
		for(int j = 0; j < s; j++)
			f << a2[i][j] << ",";
		f << std::endl;
	}
	f.close();
}

void TestFavorGammaStar()
{
	double replacement_rate = 0.02;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);
	RandomVariable *agent_interaction_rate = new UniformDistribution(0.0, 1, false);
	int communication_min_observations = 200;

	std::vector<FavorAgent *> *favor_agent_list = new std::vector<FavorAgent *>();

	FavorAgent *fa1 = new CommunicatingReciprocatingFavorAgent(0, 2, discount_factor_distro, replacement_rate, communication_min_observations, true, FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL);
	favor_agent_list->push_back(fa1);

	FavorAgent *fa2 = new CommunicatingReciprocatingFavorAgent(1, 2, discount_factor_distro, replacement_rate, communication_min_observations, true, FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL);
	favor_agent_list->push_back(fa2);

	fa1->encounterRate[1] = 0.75;
	fa2->encounterRate[0] = 0.75;

	RandomVariable *c, *w;
//	c = new UniformDistribution(0, 100);
//	c = new ExponentialDistribution(100);
//	w = new UniformDistribution(50, 150);

	//c = new UniformDistribution(0, 200);
	c = new ExponentialDistribution(100);
	w = new UniformDistribution(550, 650);
	fge = new FavorGameEnvironment(c, w, favor_agent_list);


	FavorGame fg;
	fg.c = 100;
	fg.w = w->GetExpectedValue();
	fg.a1 = fa1;
	fg.a2 = fa2;

	const int s = 32;
	double a2[s][s];
	for(int i = 0; i < s; i++)
		for(int j = 0; j < s; j++)
		{
			fg.c = 200 * (double)i / s;
			fa2->discountFactor = (double)j / s;

			double g;
			//if(ComputeGammaStar<MaxCGammaValueFavorEvaluatorRational>(fg, fa2->discountFactor, true, g))
			if(ComputeGammaStar<MaxCGammaValueFavorEvaluatorDerivative>(fg, fa2->discountFactor, true, g))
			//if(ComputeGammaStar<MaxCGammaValueFavorEvaluatorSanction>(fg, fa2->discountFactor, true, g))
				a2[i][j] = g;
			else
				a2[i][j] = -1;
		}

	std::ofstream f;
	f.open("out.csv");

	//write out headers
	f << "c\\g,";
	for(int j = 0; j < s; j++)
		f << (double)j / s << ",";
	f << std::endl;

	for(int i = 0; i < s; i++)
	{
		f << 200.0*(double)i / s << ",";	//write out row header
		for(int j = 0; j < s; j++)
			f << a2[i][j] << ",";
		f << std::endl;
	}
	f.close();
}
