#include <stdio.h>
#include <math.h>
#include <time.h>

#include <fstream>
#include <vector>
#include <algorithm>
#include <limits>

//this is awfuly long to represent infinity
#define INFINITY (std::numeric_limits<double>::infinity())

#include "RandomVariables.h"
#include "Solver.h"
#include "ObservationHistory.h"
#include "BaseAgentInteractions.h"
#include "MarketGame.h"
#include "MarketAgent.h"
#include "FavorGame.h"
#include "FavorAgent.h"
#include "FavorAgentProbabilisticReciprocity.h"
#include "RobotFinDeliveryGame.h"
#include "RobotFinDeliveryAgent.h"
#include "DiscreteFavors.h"

//global of the market game environment (used by agents)
MarketGameEnvironment *mge = NULL;
FavorGameEnvironment *fge = NULL;
RobotFinDeliveryGameEnvironment *rfdge = NULL;

void discrete_favor_main(int argc, char* argv[])
{
	DiscreteFavors df;
//	int w_vals[] = {0, 4, 10, 19, 29};
//	int c_vals[] = {0, 2, 5, 12, 23};

int w_vals[] = {0, 4, 10, 19, 29};
int c_vals[] = {0, 2, 3, 4, 5};

	for(int i = 0; i < sizeof(w_vals)/sizeof(int); i++)
	{
		df.W.push_back(w_vals[i]);
		df.C.push_back(c_vals[i]);
	}

	std::vector<int> best_strategy;

	double a1_discount_factor = .9;
	double a2_discount_factor = .8;
	
	//df.FindBestStrategyOfLengthL(a1_discount_factor, .1, a2_discount_factor, 3, 8, best_strategy);
	//df.FindBestStrategyOfLengthL(a1_discount_factor, .7, a2_discount_factor, 3, 8, best_strategy);

	df.FindBestStrategyOfLengthL(a1_discount_factor, .1, a2_discount_factor, 1, 6, best_strategy);

	for(int i = 0; i < best_strategy.size(); i++)
		printf("%d ", best_strategy[i]);

	printf("\n\na1_payoffs (left player)\n");
	std::vector< std::vector<double> > a1_payoffs;
	df.ComputeExpectedPayoffsForAllStationaryStrategies(a1_discount_factor, a1_payoffs);
	for(int i = 0; i < df.C.size(); i++)
	{
		for(int j = 0; j < df.C.size(); j++)
			printf("%.03f ", a1_payoffs[i][j]);
		printf("\n");
	}

	printf("\na2_payoffs (top player)\n");
	std::vector< std::vector<double> > a2_payoffs;
	df.ComputeExpectedPayoffsForAllStationaryStrategies(a2_discount_factor, a2_payoffs);
	for(int i = 0; i < df.C.size(); i++)
	{
		for(int j = 0; j < df.C.size(); j++)
			printf("%.03f ", a2_payoffs[j][i]);
		printf("\n");
	}

}

//arguments example: 2 u h 0 43535
void favor_main(int argc, char* argv[])
{
	if(argc < 6)
		return;
	int communication_min_observations = atoi(argv[1]);
	char connection_topology = argv[2][0];
	char discount_factor_dist = argv[3][0];
	int reciprocity_type = atoi(argv[4]);
	int rand_seed = atoi(argv[5]);

	switch(reciprocity_type)
	{
	case 0:	reciprocity_type = FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL;	break;
	case 1:	reciprocity_type = FAVOR_AGENT_RECIPROCITY_DERIVATIVE_MATCH;		break;
	case 2:	reciprocity_type = FAVOR_AGENT_RECIPROCITY_SANCTION;				break;
	}

	//to initialize from within the code
	/*
	communication_min_observations = 200;
	connection_topology = 'u';
	discount_factor_dist = 'u';
	rand_seed = 4531;
	//*/

//	init_genrand(time(NULL));
	init_genrand(531);

	double replacement_rate = 0.02;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);
	RandomVariable *agent_interaction_rate = new UniformDistribution(0.0, 1, false);
//RandomVariable *agent_interaction_rate = new UniformDistribution(1, 1, false);
	
	int num_agents = 32;//16; //5
	std::vector<FavorAgent *> *favor_agent_list = new std::vector<FavorAgent *>();
	std::vector<Agent *> *agent_list = new std::vector<Agent *>();
	agent_list->reserve(num_agents);
	for(int i = 0; i < num_agents; i++)
	{
		FavorAgent *fa;
		if(i < 800)
			fa = new CommunicatingReciprocatingFavorAgent(i, num_agents, discount_factor_distro, replacement_rate, communication_min_observations, true, FAVOR_AGENT_RECIPROCITY_INDIVIDUALLY_RATIONAL);
		else
			fa = new FavorAgentProbabilisticReciprocityAgent(i, num_agents, discount_factor_distro, replacement_rate);
//			fa = new FixedFavorAgent(i, num_agents, discount_factor_distro, replacement_rate, true);

		favor_agent_list->push_back(fa);
		agent_list->push_back(fa);
	}

	for(int i = 0; i < num_agents; i++)
	{
		double df;
		switch(discount_factor_dist)
		{
		case 'h':	df = pow((double)i / num_agents, 0.25) + .001;			break;
		case 'l':	df = 1.0 - pow((double)i / num_agents, 0.25) + .001;	break;
		case 'u':	df = (double)i / num_agents + .001;		break;
		case 's':	df = .8;	break;
		}
		if(df >= 1.0)
			df = 0.999;
		(*agent_list)[i]->discountFactor = df;
	}

	RandomVariable *c, *w;
//	c = new UniformDistribution(0, 100);
//	c = new ExponentialDistribution(100);
//	w = new UniformDistribution(50, 150);

	c = new UniformDistribution(0, 200);
	w = new UniformDistribution(550, 650);
	fge = new FavorGameEnvironment(c, w, favor_agent_list);

init_genrand(rand_seed);
	if(connection_topology == 'u')
		fge->InitializeUniformConnections(agent_interaction_rate);
	else
		fge->InitializeSmallWorldConnections(agent_interaction_rate);
	/*
	PrintConnections(agent_list);
	char aa;
	scanf("%c", &aa);
	return 0;//*/

//set rand for game payoffs
//init_genrand(23582);
//init_genrand(time(NULL));
//init_genrand(rand_seed);

	fge->RunSimulation(50);
}

//arguments example: 2 u h 43535
void market_main(int argc, char* argv[])
{
	if(argc < 5)
		return;
	int communication_min_observations = atoi(argv[1]);
	char connection_topology = argv[2][0];
	char discount_factor_dist = argv[3][0];
	int rand_seed = atoi(argv[4]);

	//to initialize from within the code
	/*
	communication_min_observations = 500;
	connection_topology = 'u';
	discount_factor_dist = 'h';
	rand_seed = 4531;
	//*/

//	init_genrand(time(NULL));
	init_genrand(4531);

	double replacement_rate = 0.01;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);
	RandomVariable *agent_interaction_rate = new UniformDistribution(0.0, 1, false);
	
	int num_agents = 5;
	std::vector<MarketAgent *> *market_agent_list = new std::vector<MarketAgent *>();
	std::vector<Agent *> *agent_list = new std::vector<Agent *>();
	agent_list->reserve(num_agents);
	for(int i = 0; i < num_agents; i++)
	{
		ADFDAgent *a = new ADFDAgent(i, num_agents, discount_factor_distro, replacement_rate, communication_min_observations);
		market_agent_list->push_back(a);
		agent_list->push_back(a);
	}

	for(int i = 0; i < num_agents; i++)
	{
		double df;
		switch(discount_factor_dist)
		{
		case 'h':	df = pow((double)i / num_agents, 0.25) + .001;			break;
		case 'l':	df = 1.0 - pow((double)i / num_agents, 0.25) + .001;	break;
		case 'u':	df = (double)i / num_agents + .001;		break;
		case 's':	df = .7;	break;
		}
		if(df >= 1.0)
			df = 0.999;
		(*agent_list)[i]->discountFactor = df;
	}

	RandomVariable *c, *m, *w;
	c = new UniformDistribution(0, 5);
	m = new UniformDistribution(10, 20);
	w = new UniformDistribution(80, 120);
	mge = new MarketGameEnvironment(c, m, w, market_agent_list);

init_genrand(rand_seed);
	if(connection_topology == 'u')
		mge->InitializeUniformConnections(agent_interaction_rate);
	else
		mge->InitializeSmallWorldConnections(agent_interaction_rate);
	/*
	PrintConnections(agent_list);
	char aa;
	scanf("%c", &aa);
	return 0;//*/

//set rand for game payoffs
//init_genrand(23582);
//init_genrand(time(NULL));
//init_genrand(rand_seed);

	mge->RunSimulation(100);
}

void robot_main(int argc, char* argv[])
{
	init_genrand(time(NULL));

	double replacement_rate = 0.01;
	RandomVariable *discount_factor_distro = new UniformDistribution(0, 1, false);

	int num_tasks = 500;
	int num_agents = 5;
	std::vector<RobotFinDeliveryAgent *> *rfd_agent_list = new std::vector<RobotFinDeliveryAgent *>();
	std::vector<Agent *> *agent_list = new std::vector<Agent *>();
	agent_list->reserve(num_agents);
	for(int i = 0; i < num_agents; i++)
	{
		RandomVariable *distance_distro = new StratifiedDiscreteUniformDistribution(1, 3, num_tasks);
		RobotFinDeliveryAgent *a;
		if(i % 2)
			a = new RobotFinDeliveryProbabilisticReciprocityAgent(i, num_agents, distance_distro, discount_factor_distro, replacement_rate);
		else
			a = new RobotFinDeliveryDiscountFactorMeasuringAgent(i, num_agents, distance_distro, discount_factor_distro, replacement_rate, 3);
		rfd_agent_list->push_back(a);
		agent_list->push_back(a);
	}

	RandomVariable *fin_number_distro = fin_number_distro = new UniformDistribution(0, 4);
	rfdge = new RobotFinDeliveryGameEnvironment(fin_number_distro, num_tasks, rfd_agent_list);

	rfdge->RunSimulation();

	//print average distance
	double ave_dist = 0.0;
	for(int i = 0; i < num_agents; i++)
		ave_dist += (*rfd_agent_list)[i]->totalTravelDistance;
	ave_dist /= num_agents;
	printf("average distance: %f", ave_dist);
}

#include "TestCode.h"

int main(int argc, char* argv[])
{
//	TestObservationHistory();

//	discrete_favor_main(argc, argv);

//	TestFavorGammaC();
//	TestFavorGammaStar();
//	favor_main(argc, argv);

	market_main(argc, argv);
//	robot_main(argc, argv);

	char a;
//	scanf("%c", &a);
	
	return 0;
}
